import asyncio
import random
import time
from collections import (
    defaultdict,
    deque
)
from hedra.distributed.env import (
    Env, 
    ReplicationEnv,
    load_env
)
from hedra.distributed.env.time_parser import TimeParser
from hedra.distributed.hooks.client_hook import client
from hedra.distributed.hooks.server_hook import server
from hedra.distributed.models.raft import (
    RaftMessage,
    HealthCheck,
    ElectionState,
    VoteResult
)
from hedra.distributed.types import Call
from hedra.distributed.models.raft.logs import Entry, NodeState
from hedra.distributed.monitoring import Monitor
from hedra.distributed.snowflake.snowflake_generator import (
    SnowflakeGenerator,
    Snowflake
)
from hedra.logging import (
    HedraLogger,
    logging_manager
)
from hedra.tools.helpers import cancel
from typing import (
    Optional, 
    Union, 
    Deque,
    Dict, 
    Tuple, 
    List,
    Any
)
from .log_queue import LogQueue


class ReplicationController(Monitor):

    def __init__(
        self,
        host: str,
        port: int,
        env: Optional[Env]=None,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        logs_directory: Optional[str]=None,
        workers: int=0,
    ) -> None:
        
        if env is None:
            env = load_env(Env)

        if logs_directory is None:
            logs_directory = env.MERCURY_SYNC_LOGS_DIRECTORY

        replication_env = load_env(ReplicationEnv) 

        super().__init__(
            host,
            port,
            env=env,
            cert_path=cert_path,
            key_path=key_path,
            workers=workers,
            logs_directory=logs_directory
        )

        self._models = [
            HealthCheck,
            RaftMessage
        ]

        self._term_number = 0
        self._term_votes = defaultdict(
            lambda: defaultdict(
                lambda: 0
            )
        )

        self._max_election_timeout = TimeParser(
            replication_env.MERCURY_SYNC_RAFT_ELECTION_MAX_TIMEOUT
        ).time

        self._min_election_timeout = max(
            self._max_election_timeout * 0.5,
            1
        )

        self._election_poll_interval = TimeParser(
            replication_env.MERCURY_SYNC_RAFT_ELECTION_POLL_INTERVAL
        ).time

        self._logs_update_poll_interval = TimeParser(
            replication_env.MERCURY_SYNC_RAFT_LOGS_UPDATE_POLL_INTERVAL
        ).time
        
        self._election_status = ElectionState.READY
        self._raft_node_status = NodeState.FOLLOWER
        self._active_election_waiter: Union[asyncio.Future, None] = None
        self._latest_election: Dict[int, int] = {}
        self._term_leaders: List[Tuple[str, int]] = []
        
        self._running = False

        self._logs = LogQueue()
        self._previous_entry_index = 0
        self._term_number = 0

        self._raft_monitor_task: Union[asyncio.Task, None] = None
        self._tasks_queue: Deque[asyncio.Task] = deque()
        self._entry_id_generator = SnowflakeGenerator(self._instance_id)

        logging_manager.logfiles_directory = logs_directory
        logging_manager.update_log_level(
            env.MERCURY_SYNC_LOG_LEVEL
        )

        self._logger = HedraLogger()
        self._logger.initialize()


        self._election_poll_interval = TimeParser(
            replication_env.MERCURY_SYNC_RAFT_ELECTION_POLL_INTERVAL
        ).time

        self._cleanup_interval = TimeParser(
            env.MERCURY_SYNC_CLEANUP_INTERVAL
        ).time

        self.registration_timeout = TimeParser(
            replication_env.MERCURY_SYNC_RAFT_REGISTRATION_TIMEOUT
        ).time

        self._pending_election_waiter: Union[asyncio.Future, None]  = None

        self._election_timeout = random.uniform(
            self._min_election_timeout,
            self._max_election_timeout
        )

        self._raft_cleanup_task: Union[asyncio.Future, None] = None
        self._election_task: Union[asyncio.Task, None] = None 
        self._active_election = False
        
    async def start(self):

        await self._logger.filesystem.aio.create_logfile(f'hedra.distributed.{self._instance_id}.log')
        self._logger.filesystem.create_filelogger(f'hedra.distributed.{self._instance_id}.log')

        await self._logger.distributed.aio.info(f'Starting server for node - {self.host}:{self.port} - with id - {self._instance_id}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Starting server for node - {self.host}:{self.port} - with id - {self._instance_id}')

        await self.start_server()

        self._instance_ids[(self.host, self.port)] = Snowflake.parse(
            self._entry_id_generator.generate()
        ).instance

        boot_wait = random.uniform(0.1, self.boot_wait * self._initial_expected_nodes)
        await asyncio.sleep(boot_wait)

    async def register(
        self,
        host: str,
        port: int
    ):
    
        await self._logger.distributed.aio.info(f'Initializing node - {self.host}:{self.port} - with id - {self._instance_id}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Initializing node - {self.host}:{self.port} - with id - {self._instance_id}')
   
        self.bootstrap_host = host
        self.bootstrap_port = port
        self.status = 'healthy'

        await self._register_initial_node()
        await self._run_registration()

        self._running = True

        self._healthcheck_task = asyncio.create_task(
            self.start_health_monitor()
        )

        self._cleanup_task = asyncio.create_task(
            self.cleanup_pending_checks()
        )

        self._udp_sync_task = asyncio.create_task(
            self._run_udp_state_sync()
        )

        self._tcp_sync_task = asyncio.create_task(
            self._run_tcp_state_sync()
        )

        boot_wait = random.uniform(0.1, self.boot_wait * self._initial_expected_nodes)
        await asyncio.sleep(boot_wait)

        if self._term_number == 0:
            self._election_status = ElectionState.ACTIVE
            await self.run_election()

        self._raft_cleanup_task = asyncio.create_task(
            self._cleanup_pending_raft_tasks()
        )

        self._raft_monitor_task = asyncio.create_task(
            self._run_raft_monitor()
        )

        self.status = 'healthy'

    async def _run_registration(self):

        last_registered_count = -1
        poll_timeout = self.registration_timeout * self._initial_expected_nodes
        
        while self._check_all_nodes_registered() is False:

            monitors = [
                address for address in self._node_statuses.keys()
            ]

            active_nodes_count = len(monitors)
            registered_count = self._calculate_all_registered_nodes()

            if registered_count > last_registered_count:
                await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - reporting - {registered_count}/{self._initial_expected_nodes} - as fully registered')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - reporting - {registered_count}/{self._initial_expected_nodes} - as fully registered')

                last_registered_count = registered_count

            if active_nodes_count > 0:

                for host, port in monitors:
                    self._tasks_queue.append(
                        asyncio.create_task(
                            asyncio.wait_for(
                                self._submit_registration(
                                    host,
                                    port
                                ),
                                timeout=poll_timeout
                            )
                        )
                    )

                    await asyncio.sleep(self._poll_interval)

            await asyncio.sleep(self._poll_interval)

        registered_count = self._calculate_all_registered_nodes()

        await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - reporting - {registered_count}/{self._initial_expected_nodes} - as fully registered')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - reporting - {registered_count}/{self._initial_expected_nodes} - as fully registered')

    def _calculate_all_registered_nodes(self) -> int:
        self._registered_counts[(self.host, self.port)] = len(self._instance_ids)
        return len([
            count for count in self._registered_counts.values() if count == self._initial_expected_nodes
        ])

    def _check_all_nodes_registered(self) -> bool:
        return self._calculate_all_registered_nodes() == self._initial_expected_nodes

    async def _submit_registration(
        self,
        host: str,
        port: int
    ):
        shard_id, response = await self.submit_registration(
            host,
            port
        )

        if isinstance(response, HealthCheck):
            source_host = response.source_host
            source_port = response.source_port

            not_self = self._check_is_not_self(
                source_host,
                source_port
            )

            self._instance_ids[(source_host, source_port)] = Snowflake.parse(shard_id).instance

            if not_self:
                self._node_statuses[(source_host, source_port)] = 'healthy'

            self._registered_counts[(source_host, source_port)] = max(
                response.registered_count,
                self._registered_counts[(source_host, source_port)]
            )

    @server()
    async def receive_vote_request(
        self,
        shard_id: int,
        raft_message: RaftMessage
    ) -> Call[RaftMessage]:

        source_host = raft_message.source_host
        source_port = raft_message.source_port

        term_number = raft_message.term_number

        elected_host: Union[str, None] = None
        elected_port: Union[int, None] = None

        if term_number > self._term_number:
            # The requesting node is ahead. They're elected the leader by default.
            elected_host = source_host
            elected_port = source_port

        elif term_number == self._term_number and self._raft_node_status != NodeState.LEADER:
            # The term numbers match, we can choose a candidate.

            elected_host, elected_port = self._get_max_instance_id()

        else:

            leader_host, leader_port = self._term_leaders[-1]

            return RaftMessage(
                host=source_host,
                port=source_port,
                source_host=self.host,
                source_port=self.port,
                elected_leader=(
                    leader_host,
                    leader_port
                ),
                status=self.status,
                error='Election request term cannot be less than current term.',
                vote_result=VoteResult.REJECTED,
                raft_node_status=self._raft_node_status,
                term_number=self._term_number
            )
            
        vote_result = VoteResult.REJECTED

        if elected_host == source_host and elected_port == source_port:
            vote_result = VoteResult.ACCEPTED

        return RaftMessage(
            host=source_host,
            port=source_port,
            source_host=self.host,
            source_port=self.port,
            elected_leader=(
                elected_host,
                elected_port
            ),
            status=self.status,
            vote_result=vote_result,
            raft_node_status=self._raft_node_status,
            term_number=term_number
        )
    
    @server()
    async def receive_log_update(
        self,
        shard_id: int,
        message: RaftMessage
    ) -> Call[RaftMessage]:
        
        entries_count = len(message.entries)

        if entries_count < 1:
            return RaftMessage(
                host=message.host,
                port=message.port,
                source_host=self.host,
                source_port=self.port,
                status=self.status,
                term_number=self._term_number,
                election_status=self._election_status,
                raft_node_status=self._raft_node_status
            )
        
        # We can use the Snowflake ID to sort since all records come from the 
        # leader.
        entries: List[Entry] = list(
            sorted(
                message.entries,
                key=lambda entry: Snowflake.parse(
                    entry.entry_id
                ).timestamp
            )
        )

        last_entry = entries[-1]

        leader_host = last_entry.leader_host
        leader_port = last_entry.leader_port
        
        try:


            if message.term_number > self._term_number:

                self._tasks_queue.append(
                    asyncio.create_task(
                        self._cancel_election(message)
                    )
                )
                
                amount_behind = max(
                    message.term_number - self._term_number - 1,
                    0
                )

                last_entry = entries[-1]

                leader_host = last_entry.leader_host
                leader_port = last_entry.leader_port

                self._term_number = message.term_number

                for _ in range(amount_behind):
                    self._term_leaders.append((
                        None,
                        None
                    ))

                await self._logger.distributed.aio.info(f'Term number for source - {self.host}:{self.port} - was updated to - {self._term_number} - and leader was updated to - {leader_host}:{leader_port}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Term number for source - {self.host}:{self.port} - was updated to - {self._term_number} - and leader was updated to - {leader_host}:{leader_port}')


                self._term_leaders.append((
                    leader_host,
                    leader_port
                ))

                self._election_status = ElectionState.READY
                self._raft_node_status = NodeState.FOLLOWER

                return RaftMessage(
                    host=message.source_host,
                    port=message.source_port,
                    source_host=self.host,
                    source_port=self.port,
                    elected_leader=(
                        leader_host,
                        leader_port
                    ),
                    status=self.status,
                    error='Election request term cannot be less than current term.',
                    vote_result=VoteResult.REJECTED,
                    raft_node_status=self._raft_node_status,
                    term_number=self._term_number
                )


            source_host = message.source_host
            source_port = message.source_port

            if message.failed_node and self._suspect_tasks.get(
                message.failed_node
            ):
                
                node_host, node_port = message.failed_node
                
                self._tasks_queue.append(
                    asyncio.create_task(
                        self._cancel_suspicion_probe(
                            node_host,
                            node_port
                        )
                    )
                )

                await self._logger.distributed.aio.debug(f'Node - {node_host}:{node_port} - submitted healthy status to source - {self.host}:{self.port} - and is no longer suspect')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {node_host}:{node_port} - submitted healthy status to source - {self.host}:{self.port} - and is no longer suspect')


            if self._suspect_tasks.get((
                source_host,
                source_port
            )):
                self._tasks_queue.append(
                    asyncio.create_task(
                        self._cancel_suspicion_probe(
                            source_host,
                            source_port
                        )
                    )
                )

                await self._logger.distributed.aio.debug(f'Node - {source_host}:{source_port} - submitted healthy status to source - {self.host}:{self.port} - and is no longer suspect')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {source_host}:{source_port} - submitted healthy status to source - {self.host}:{self.port} - and is no longer suspect')

            error = self._logs.update(entries)

            self._local_health_multipliers[(source_host, source_port)] = self._reduce_health_multiplier(
                source_host,
                source_port
            )

            if isinstance(error, Exception):

                return RaftMessage(
                    host=message.source_host,
                    port=message.source_port,
                    source_host=self.host,
                    source_port=self.port,
                    status=self.status,
                    raft_node_status=self._raft_node_status,
                    error=str(error),
                    elected_leader=(
                        leader_host,
                        leader_port
                    ),
                    term_number=self._term_number
                )

            return RaftMessage(
                host=message.source_host,
                port=message.source_port,
                source_host=self.host,
                source_port=self.port,
                status=self.status,
                elected_leader=(
                    leader_host,
                    leader_port
                ),
                term_number=self._term_number,
                raft_node_status=self._raft_node_status,
                received_timestamp=self._logs.last_timestamp
            )
        
        except Exception as rpc_error:
            return RaftMessage(
                host=message.source_host,
                port=message.source_port,
                source_host=self.host,
                source_port=self.port,
                status=self.status,
                raft_node_status=self._raft_node_status,
                error=str(rpc_error),
                elected_leader=(
                    leader_host,
                    leader_port
                ),
                term_number=self._term_number
            )

    @server()
    async def receive_forwarded_entries(
        self,
        shard_id: int,
        message: RaftMessage
    ) -> Call[RaftMessage]:
        
        if self._raft_node_status == NodeState.LEADER and message.entries:
            
            entries = message.entries

            entries.append(
                Entry.from_data(
                    entry_id=self._entry_id_generator.generate(),
                    leader_host=self.host,
                    leader_port=self.port,
                    term=self._term_number,
                    data={
                        'key': 'logs_update',
                        'value': f'Node - {self.host}:{self.port} - submitted log update',
                    }
                )
            )

            self._tasks_queue.append(
                asyncio.create_task(
                    self._submit_logs_to_members(entries)
                )
            )

        return RaftMessage(
            host=message.host,
            port=message.port,
            source_host=self.host,
            source_port=self.port,
            status=self.status,
            term_number=self._term_number,
            raft_node_status=self._raft_node_status,
            received_timestamp=self._logs.last_timestamp
        )
        
    @server()
    async def receive_failure_notification(
        self,
        shard_id: int,
        message: RaftMessage
    ) -> Call[RaftMessage]:
        
        try:
            failed_node = message.failed_node
            host, port = failed_node

            not_self = self._check_is_not_self(
                host,
                port
            )


            if not_self and self._election_status == ElectionState.READY and failed_node not in self.failed_nodes:


                self.failed_nodes.append((
                    host,
                    port,
                    time.monotonic()
                ))

                self._node_statuses[failed_node] = 'failed'

                self._election_status = ElectionState.ACTIVE
                
                self._tasks_queue.append(
                    asyncio.create_task(
                        self.run_election(
                            failed_node=failed_node
                        )
                    )
                )

            return RaftMessage(
                host=message.host,
                port=message.port,
                source_host=self.host,
                source_port=self.port,
                status=self.status,
                term_number=self._term_number,
                raft_node_status=self._raft_node_status,
                received_timestamp=self._logs.last_timestamp
            )
        
        except Exception:
            pass

    @client('receive_vote_request')
    async def request_vote(
        self,
        host: str,
        port: int
    ) -> Call[RaftMessage]:
        return RaftMessage(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            status=self.status,
            term_number=self._term_number,
            raft_node_status=self._raft_node_status
        )
    
    @client('receive_log_update')
    async def submit_log_update(
        self,
        host: str,
        port: int,
        entries: List[Entry],
        failed_node: Optional[Tuple[str, int]]=None
    ) -> Call[RaftMessage]:
        return RaftMessage(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            status=self.status,
            term_number=self._term_number,
            raft_node_status=self._raft_node_status,
            failed_node=failed_node,
            entries=entries
        )
    
    @client('receive_forwarded_entries')
    async def forward_entries_to_leader(
        self,
        host: str,
        port: int,
        entries: List[Entry]
    ) -> Call[RaftMessage]:
        return RaftMessage(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            status=self.status,
            term_number=self._term_number,
            raft_node_status=self._raft_node_status,
            entries=entries
        )
    
    @client('receive_failure_notification')
    async def submit_failure_notification(
        self,
        host: str,
        port: int,
        failed_node: Tuple[str, int]
    ) -> Call[RaftMessage]:
        return RaftMessage(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            status=self.status,
            term_number=self._term_number,
            raft_node_status=self._raft_node_status,
            failed_node=failed_node
        )
            
    async def _start_suspect_monitor(self):
        suspect_host, suspect_port = await super()._start_suspect_monitor()

        node_status = self._node_statuses.get((suspect_host, suspect_port))

        failed_node = (suspect_host, suspect_port)

        if self._election_status == ElectionState.READY and node_status == 'failed' and failed_node not in self.failed_nodes:

            self.failed_nodes.append((
                suspect_host,
                suspect_port,
                time.monotonic()
            ))

            self._election_status = ElectionState.ACTIVE

            await self.notify_of_failed_node(failed_node=failed_node)
            await self.run_election(failed_node=failed_node)

    async def push_entries(
        self,
        entries: List[Dict[str, Any]]
    ) -> List[RaftMessage]:

        entries.append({
            'key': 'logs_update',
            'value': f'Node - {self.host}:{self.port} - submitted log update',
        })

        entries = self._convert_data_to_entries(entries)
        entries_count = len(entries)

        if self._raft_node_status == NodeState.LEADER:
            results = await self._submit_logs_to_members(entries)

            results_count = len(results)

            await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - pushed - {entries_count} - entries to - {results_count} - members')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - pushed - {entries_count} - entries to - {results_count} - members')

            return results

        else:

            try:

                current_leader_host, current_leader_port = self._term_leaders[-1]

                result = await asyncio.wait_for(
                    self.forward_entries_to_leader(
                        current_leader_host,
                        current_leader_port,
                        entries
                    ),
                    timeout=self._calculate_current_timeout(
                        current_leader_host,
                        current_leader_port
                    )
                )

                await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - forwarded - {entries_count} - entries to leader at - {current_leader_host}:{current_leader_port}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - forwarded - {entries_count} - entries to leader at - {current_leader_host}:{current_leader_port}')

                return [
                    result
                ]
                
            except Exception as forward_error:
                
                await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - encountered error - {str(forward_error)} - out forwarding - {entries_count} - entries to leader at - {current_leader_host}:{current_leader_port}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - encountered error - {str(forward_error)} - out forwarding - {entries_count} - entries to leader at - {current_leader_host}:{current_leader_port}')
        
                return [
                    RaftMessage(
                        host=current_leader_host,
                        port=current_leader_port,
                        source_host=self.host,
                        source_port=self.port,
                        elected_leader=(
                            current_leader_host,
                            current_leader_port
                        ),
                        error=str(forward_error),
                        raft_node_status=self._raft_node_status,
                        status=self.status,
                        term_number=self._term_number
                    )
                ]
            
    def submit_entries(
        self,
        entries: List[Dict[str, Any]]
    ):
        self._tasks_queue.append(
            asyncio.create_task(
                self.push_entries(entries)
            )
        )

    def _convert_data_to_entries(
        self,
        entries: List[Dict[str, Any]]
    ) -> List[Entry]:
        
        current_leader_host, current_leader_port = self._term_leaders[-1]

        entries = [
            Entry.from_data(
                self._entry_id_generator.generate(),
                current_leader_host,
                current_leader_port,
                self._term_number,
                entry
            ) for entry in entries
        ]

        return entries
    
    def _get_max_instance_id(self):

        nodes = [
            address for address, status in self._node_statuses.items() if status == 'healthy'
        ]

        nodes.append((
            self.host,
            self.port
        ))

        instance_address_id_pairs = list(
            sorted(
                nodes,
                key=lambda instance: self._instance_ids.get(
                    instance,
                    self._instance_id
                )
            )
        )

        if len(instance_address_id_pairs) > 0:

            max_instance = instance_address_id_pairs[-1]
            elected_host, elected_port = max_instance

        else:
            
            elected_host = self.host
            elected_port = self.port

        return elected_host, elected_port

    async def _cancel_election(
        self,
        message: RaftMessage
    ):
        self._election_status = ElectionState.READY
        self._term_number = message.term_number

        if self._election_task:
            await cancel(self._election_task)
            self._election_task = None

            await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - election for term - {self._term_number} - was cancelled due to leader reporting for term')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - election for term - {self._term_number} - was cancelled due to leader reporting for term')

    async def _update_logs(
        self,
        host: str,
        port: int,
        entries: List[Entry],
        failed_node: Optional[Tuple[str, int]]=None,
    ) -> Union[
        Tuple[int, RaftMessage],
        None
    ]:
        shard_id: Union[int, None] = None
        update_response: Union[RaftMessage, None] = None

        await self._logger.distributed.aio.debug(f'Running UDP logs update for node - {host}:{port} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Running UDP logs update for node - {host}:{port} - for source - {self.host}:{self.port}')
        
        for idx in range(self._poll_retries):

            try:

                response = await asyncio.wait_for(
                    self.submit_log_update(
                        host,
                        port,
                        entries,
                        failed_node=failed_node
                    ),
                    timeout=self._calculate_current_timeout(
                        host,
                        port
                    )
                )

                shard_id, update_response = response
                source_host, source_port = update_response.source_host, update_response.source_port

                not_self = self._check_is_not_self(
                    source_host,
                    source_port
                )

                if not_self:
                    self._node_statuses[(source_host, source_port)] = update_response.status

                self._local_health_multipliers[(host, port)] = self._reduce_health_multiplier(
                    host,
                    port
                )

                return shard_id, update_response
            
            except Exception:

                self._local_health_multipliers[(host, port)] = self._increase_health_multiplier(
                    host,
                    port
                )

        check_host = host
        check_port = port

        node_status = self._node_statuses.get((
            check_host, 
            check_port
        )) 

        not_self = self._check_is_not_self(
            check_host,
            check_port
        )

        if not_self and update_response is None and node_status == 'healthy':

            await self._logger.distributed.aio.debug(f'Node - {check_host}:{check_port} - failed to respond over - {self._poll_retries} - retries and is now suspect for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {check_host}:{check_port} - failed to respond over - {self._poll_retries} - retries and is now suspect for source - {self.host}:{self.port}')

            self._node_statuses[(check_host, check_port)] = 'suspect'

            self._suspect_nodes.append((
                check_host,
                check_port
            ))

            self._suspect_tasks[(host, port)] = asyncio.create_task(
                self._start_suspect_monitor()
            )

        else:

            await self._logger.distributed.aio.debug(f'Node - {check_host}:{check_port} - responded on try - {idx}/{self._poll_retries} - for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {check_host}:{check_port} - responded on try - {idx}/{self._poll_retries} - for source - {self.host}:{self.port}')

    def _calculate_current_timeout(
        self,
        host: str,
        port: int
    ):

        modifier = max(
            len([
                address for address, status in self._node_statuses.items() if status == 'healthy' 
            ]),
            self._initial_expected_nodes
        )

        return self._poll_timeout * (self._local_health_multipliers[(host, port)] + 1) * modifier
    
    async def notify_of_failed_node(
        self,
        failed_node: Tuple[str, int]
    ):
        monitors = [
            address for address, status in self._node_statuses.items() if status == 'healthy' and address != failed_node
        ]

        responses: List[
            Union[
                Tuple[int, RaftMessage],
                Exception
            ]
        ] = await asyncio.gather(*[
            asyncio.wait_for(
                self.submit_failure_notification(
                    host,
                    port,
                    failed_node
                ),
                timeout=self._calculate_current_timeout(
                    host,
                    port
                )
            ) for host , port in monitors
        ], return_exceptions=True)

        for response in responses:
            if isinstance(response, Exception):
                raise response
    
    
    async def run_election(
        self,
        failed_node: Optional[Tuple[str, int]]=None
    ):

        # Trigger new election
        next_term = self._term_number + 1
        self._raft_node_status = NodeState.CANDIDATE

        await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - Running election for term - {next_term}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - Running election for term - {next_term}')

        elected_host, elected_port = self._get_max_instance_id()
        self._term_leaders.append((elected_host, elected_port))

        if elected_host == self.host and elected_port == self.port:

            await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - was elected as leader for term - {next_term}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - was elected as leader for term - {next_term}')


            self._raft_node_status = NodeState.LEADER
            self._term_number += 1

            members: List[Tuple[str, int]] = [
                address for address, status in self._node_statuses.items() if status == 'healthy'
            ]

            members = list(set(members))

            self._logs.update([
                Entry.from_data(
                    entry_id=self._entry_id_generator.generate(),
                    leader_host=self.host,
                    leader_port=self.port,
                    term=self._term_number,
                    data={
                        'key': 'election_update',
                        'value': f'Election complete! Elected - {self.host}:{self.port}'
                    }
                )
            ])

            members: List[Tuple[str, int]] = [
                address for address, status in self._node_statuses.items() if status == 'healthy'
            ]

            latest_logs = self._logs.latest()

            await asyncio.gather(*[
                asyncio.wait_for(
                    self._update_logs(
                        host,
                        port,
                        latest_logs,
                        failed_node=failed_node
                    ),
                    timeout=self._calculate_current_timeout(
                        host,
                        port
                    )
                ) for host, port in members
            ], return_exceptions=True)

        else:

            self._raft_node_status = NodeState.FOLLOWER

            await self._logger.distributed.aio.info(f'Source - {self.host}:{self.port} - failed to receive majority votes and is reverting to a follower for term - {self._term_number}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - failed to receive majority votes and is reverting to a follower for term - {self._term_number}')

        if self._term_number > next_term:
            self._term_number = next_term
        
        self._election_status = ElectionState.READY
        
        return

    async def _run_raft_monitor(self):

        while self._running:

            if self._raft_node_status == NodeState.LEADER:

                self._tasks_queue.append(
                    asyncio.create_task(
                        self._submit_logs_to_members([
                            Entry.from_data(
                                entry_id=self._entry_id_generator.generate(),
                                leader_host=self.host,
                                leader_port=self.port,
                                term=self._term_number,
                                data={
                                    'key': 'logs_update',
                                    'value': f'Node - {self.host}:{self.port} - submitted log update',
                                }
                            )
                        ])
                    )
                )

            await asyncio.sleep(
                self._logs_update_poll_interval * self._initial_expected_nodes
            )

    async def _submit_logs_to_members(
        self,
        entries: List[Entry]
    ) -> List[RaftMessage]:
         
        members: List[Tuple[str, int]] = [
            address for address, status in self._node_statuses.items() if status == 'healthy'
        ]

        self._logs.update(entries)

        latest_logs = self._logs.latest()

        results: List[Tuple[
            int,
            RaftMessage
        ]] = await asyncio.gather(*[
            asyncio.create_task(
                self._update_logs(
                    host,
                    port,
                    latest_logs
                )
            ) for host, port in members
        ])

        self._logs.commit()

        return results
    
    async def _cleanup_pending_raft_tasks(self):

        await self._logger.distributed.aio.debug(f'Running cleanup for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Running cleanup for source - {self.host}:{self.port}')

        while self._running:

            pending_count = 0

            for pending_task in list(self._tasks_queue):
                if pending_task.done() or pending_task.cancelled():
                    try:
                        await pending_task

                    except Exception:
                        pass

                    self._tasks_queue.remove(pending_task)
                    pending_count += 1

            await self._logger.distributed.aio.debug(f'Cleaned up - {pending_count} - for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Cleaned up - {pending_count} - for source - {self.host}:{self.port}')

            await asyncio.sleep(self._logs_update_poll_interval)
            self._logs.prune()

    async def leave(self):
        await self._logger.distributed.aio.debug(f'Shutdown requested for RAFT source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Shutdown requested for RAFT source - {self.host}:{self.port}')

        await cancel(self._raft_monitor_task)
        await cancel(self._raft_cleanup_task)

        if self._election_task:
            await cancel(self._election_task)
            self._election_task = None

        await self._submit_leave_requests()
        await self._shutdown()

        await self._logger.distributed.aio.debug(f'Shutdown complete for RAFT source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Shutdown complete for RAFT source - {self.host}:{self.port}')



import os
from pydantic import BaseModel

from hedra.reporting.types.common.types import ReporterTypes


class SQLiteConfig(BaseModel):
    path: str=f'{os.getcwd()}/results.db'
    events_table: str='events'
    metrics_table: str='metrics'
    experiments_table: str='experiments'
    streams_table: str='streams'
    system_metrics_table: str='system_metrics'
    reporter_type: ReporterTypes=ReporterTypes.SQLite

    class Config:
        arbitrary_types_allowed = True


from __future__ import annotations
import sys
import asyncio
import time
import signal
import functools
import inspect
from enum import Enum
from os import get_terminal_size
from typing import (
    List, 
    Mapping, 
    Any, 
    Dict, 
    Coroutine
)
from asyncio import Task
from yaspin.core import Yaspin, to_unicode
from yaspin.spinners import Spinners
from aiologger.levels import LogLevel
from aiologger.formatters.base import Formatter
from hedra.logging.spinner import ProgressText
from .async_logger import AsyncLogger
from .logger_types import LoggerTypes


class LoggerMode(Enum):
    CONSOLE='console'
    SYSTEM='system'


async def default_handler(signame: str,spinner: AsyncSpinner):  # pylint: disable=unused-argument
    """Signal handler, used to gracefully shut down the ``spinner`` instance
    when specified signal is received by the process running the ``spinner``.

    ``signum`` and ``frame`` are mandatory arguments. Check ``signal.signal``
    function for more details.
    """
    await spinner.fail()
    await spinner.stop()


class AsyncSpinner(Yaspin):

    def __init__(
        self, 
        logger_name: str=None,
        logger_type: LoggerTypes=LoggerTypes.SPINNER,
        log_level: LogLevel=LogLevel.NOTSET,
        logger_enabled: bool=True,
        spinner: Spinners=None, 
        text: ProgressText=None, 
        color: str=None, 
        on_color: str=None, 
        attrs: List[str]=None, 
        reversal: bool=False, 
        side: str="left", 
        sigmap: Dict[signal.Signals, Coroutine]=None, 
        timer: bool=False,
        enabled: bool=True
    ):
        super().__init__(
            spinner, 
            text, 
            color, 
            on_color, 
            attrs, 
            reversal, 
            side, 
            sigmap, 
            timer
        )

        self.logger: AsyncLogger = AsyncLogger(
            logger_name=logger_name,
            logger_type=logger_type,
            log_level=log_level,
            logger_enabled=logger_enabled
        )

        self.logger.initialize('%(message)s')
        self.display = text

        self.enabled = enabled
        self.logger_enabled = True
        self.logger_mode = LoggerMode.CONSOLE

        self._stdout_lock = asyncio.Lock()
        self._loop = asyncio.get_event_loop()

    @property
    def console(self):
        if self.logger_mode == LoggerMode.SYSTEM:
            for handler in self.logger.handlers:
                handler.formatter = Formatter('%(message)s')

            self.logger_mode = LoggerMode.CONSOLE

        return self

    @property
    def system(self):
        if self.logger_mode == LoggerMode.CONSOLE:
            for handler in self.logger.handlers:
                handler.formatter = Formatter(
                        '%(asctime)s - %(name)s - %(levelname)s - %(module)s:%(funcName)s:%(lineno)d - %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%S.%Z'
                )

            self.logger_mode = LoggerMode.SYSTEM

        return self

    def append_message(self, message: str) -> Coroutine[None]:
        return self.display.append_cli_message(message)

    def set_default_message(self, message: str) -> Coroutine[None]:
        return self.display.clear_and_replace(message)

    def set_message_at(self, message_index: int, message:str) -> None:
        if message_index < len(self.display.cli_messages):
            self.display.cli_messages[message_index] = message

    def finalize(self):
        self.display.finalized = True

    def group_finalize(self):
        self.display.group_finalized = True

    async def debug(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.debug(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()

            return log_result

    async def info(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:

        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.info(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            
            return log_result

    async def warning(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.warning(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def warn(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        if self.logger_enabled:  
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.warn(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def error(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.error(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def critical(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.critical(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result
    
    async def fatal(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.fatal(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def __aenter__(self):

        if self.logger_enabled:
            self.display.group_timer.reset()
            await self.start()

        return self

    async def __aexit__(self, exc_type, exc_val, traceback):
        # Avoid stop() execution for the 2nd time

        enabled = self.enabled and self.logger_enabled

        if enabled and self._spin_thread.done() is False and self._spin_thread.cancelled() is False:
            await self.stop()
            
        return False  # nothing is handled

    def __call__(self, fn):
        @functools.wraps(fn)
        async def inner(*args, **kwargs):
            async with self:

                if inspect.iscoroutinefunction(fn):
                    return await fn(*args, **kwargs)

                else:
                    return fn(*args, **kwargs)

        return inner

    async def start(self):
        if self.enabled:

            if self._sigmap:
                self._register_signal_handlers()
            
            await self._hide_cursor()
            self._start_time = time.time()
            self._stop_time = None  # Reset value to properly calculate subsequent spinner starts (if any)  # pylint: disable=line-too-long
            self._stop_spin = asyncio.Event()
            self._hide_spin = asyncio.Event()
            try:
                self._spin_thread = asyncio.create_task(self._spin())
            finally:
                # Ensure cursor is not hidden if any failure occurs that prevents
                # getting it back
                await self._show_cursor()

            self.display.start_cli_tasks()

    async def stop(self):
        if self.enabled:
            self._stop_time = time.time()

            if self._dfl_sigmap:
                # Reset registered signal handlers to default ones
                self._reset_signal_handlers()

            if self._spin_thread:
                self._stop_spin.set()
                await self._spin_thread

            await self._clear_line()
            await self.display.stop_cli_tasks()
            await self._show_cursor()

    async def hide(self):
        """Hide the spinner to allow for custom writing to the terminal."""
        thr_is_alive = self._spin_thread and (self._spin_thread.done() is False and self._spin_thread.cancelled() is False)

        if thr_is_alive and not self._hide_spin.is_set():
            
            # set the hidden spinner flag
            self._hide_spin.set()
            await self._clear_line()

            # flush the stdout buffer so the current line
            # can be rewritten to
            await self._loop.run_in_executor(
                None,
                sys.stdout.flush
            )  

    async def show(self):
        """Show the hidden spinner."""
        thr_is_alive = self._spin_thread and (self._spin_thread.done() is False and self._spin_thread.cancelled() is False)

        if thr_is_alive and self._hide_spin.is_set():
            
            # clear the hidden spinner flag
            self._hide_spin.clear()

            # clear the current line so the spinner is not appended to it
            await self._clear_line()      

    async def write(self, text):
        if self.logger_enabled:
            """Write text in the terminal without breaking the spinner."""
            # similar to tqdm.write()
            # https://pypi.python.org/pypi/tqdm#writing-messages
            await self._stdout_lock.acquire()
            await self._clear_line()

            if isinstance(text, (str, bytes)):
                _text = to_unicode(text)
            else:
                _text = str(text)

            # Ensure output is Unicode
            assert isinstance(_text, str)

            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                
            )

            self._cur_line_len = 0    
            self._stdout_lock.release()  

    async def ok(self, text="OK"):
        if self.enabled:
            await self.display.stop_cli_tasks()
            """Set Ok (success) finalizer to a spinner."""
            _text = text if text else "OK"
            await self._freeze(_text)

    async def fail(self, text="FAIL"):
        if self.enabled:
            await self.display.stop_cli_tasks()
            """Set fail finalizer to a spinner."""
            _text = text if text else "FAIL"
            await self._freeze(_text)

    async def _freeze(self, final_text):
        """Stop spinner, compose last frame and 'freeze' it."""
        text = to_unicode(final_text)
        self._last_frame = self._compose_out(text, mode="last")

        # Should be stopped here, otherwise prints after
        # self._freeze call will mess up the spinner
        await self.stop()
        

        await self._loop.run_in_executor(
            None,
            sys.stdout.write,
            self._last_frame
        )

        self._cur_line_len = 0  

    async def _spin(self):
        while not self._stop_spin.is_set():

            if self._hide_spin.is_set():
                # Wait a bit to avoid wasting cycles
                await asyncio.sleep(self._interval)
                continue
            
            await self._stdout_lock.acquire()
            terminal_size = await self._loop.run_in_executor(
                None,
                get_terminal_size
            )

            terminal_width = terminal_size[0]

            # Compose output
            spin_phase = next(self._cycle)
            out = self._compose_out(spin_phase)

            if len(out) > terminal_width:
                out = f'{out[:terminal_width-1]}...'

            # Write
               
            await self._clear_line()
            
            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                out
            )

            await self._loop.run_in_executor(
                None,
                sys.stdout.flush
            )

            self._cur_line_len = max(self._cur_line_len, len(out))

            # Wait
            try:
                await asyncio.wait_for(self._stop_spin.wait(), timeout=self._interval)

            except asyncio.TimeoutError:
                pass

            self._stdout_lock.release()

    async def _clear_line(self):
        if sys.stdout.isatty():
            # ANSI Control Sequence EL does not work in Jupyter
            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                "\r\033[K"
            )

        else:
            fill = " " * self._cur_line_len
            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                sys.stdout.write,
                f"\r{fill}\r"
            )
            
    @staticmethod
    async def _show_cursor():
        loop = asyncio.get_event_loop()
        if sys.stdout.isatty():
            # ANSI Control Sequence DECTCEM 1 does not work in Jupyter
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                "\033[?25h"
            )

            await loop.run_in_executor(
                None,
                sys.stdout.flush
            )

    @staticmethod
    async def _hide_cursor():
        loop = asyncio.get_event_loop()
        if sys.stdout.isatty():
            # ANSI Control Sequence DECTCEM 1 does not work in Jupyter
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                "\033[?25l"
            )

            await loop.run_in_executor(
                None,
                sys.stdout.flush
            )

    def _register_signal_handlers(self):
        # SIGKILL cannot be caught or ignored, and the receiving
        # process cannot perform any clean-up upon receiving this
        # signal.
        if signal.SIGKILL in self._sigmap:
            raise ValueError(
                "Trying to set handler for SIGKILL signal. "
                "SIGKILL cannot be caught or ignored in POSIX systems."
            )

        for sig, sig_handler in self._sigmap.items():
            # A handler for a particular signal, once set, remains
            # installed until it is explicitly reset. Store default
            # signal handlers for subsequent reset at cleanup phase.
            dfl_handler = signal.getsignal(sig)
            self._dfl_sigmap[sig] = dfl_handler

            # ``signal.SIG_DFL`` and ``signal.SIG_IGN`` are also valid
            # signal handlers and are not callables.
            if callable(sig_handler):
                # ``signal.signal`` accepts handler function which is
                # called with two arguments: signal number and the
                # interrupted stack frame. ``functools.partial`` solves
                # the problem of passing spinner instance into the handler
                # function.
                sig_handler = functools.partial(sig_handler, spinner=self)

            self._loop.add_signal_handler(getattr(signal, sig.name),
                            lambda signame=sig.name: asyncio.create_task(sig_handler(self)))

    def _reset_signal_handlers(self):
        for sig, sig_handler in self._dfl_sigmap.items():
            self._loop.add_signal_handler(getattr(signal, sig.name),
                            lambda signame=sig.name: asyncio.create_task(sig_handler(signame, self)))




from __future__ import annotations
import asyncio
import functools
import json
import psutil
import uuid
import os
import signal
import re
import time
from pathlib import Path
from typing import List, TextIO, Dict, Union
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from .json_config import JSONConfig


has_connector = True


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop, 
    events_file: TextIO
): 
    try:
        events_file.close()
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


class JSON:

    def __init__(self, config: JSONConfig) -> None:
        self.events_filepath = config.events_filepath
        self.metrics_filepath = config.metrics_filepath
        self.experiments_filepath = config.experiments_filepath
        self.streams_filepath = config.streams_filepath
        self.system_metrics_filepath = config.system_metrics_filepath
        
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop: asyncio.AbstractEventLoop = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.events_file: TextIO = None
        self.experiments_file: TextIO = None
        self.metrics_file: TextIO = None
        self.streams_file: TextIO = None
        self.system_metrics_file: TextIO = None

        self.write_mode = 'w' if config.overwrite else 'a'
        self.pattern = re.compile("_copy[0-9]+")

    async def connect(self):
        self._loop = asyncio._get_running_loop()
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping connect')
        
        original_filepath = Path(self.events_filepath)
        
        directory = original_filepath.parent
        filename = original_filepath.stem

        events_file_timestamp =time.time()
        self.events_filepath = os.path.join(
            directory,
            f'{filename}_{events_file_timestamp}.json'
        )

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        if self.system_metrics_file is None:
            self.system_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.system_metrics_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.system_metrics_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Session System Metrics to file - {self.system_metrics_filepath}')

        metrics_sets: Dict[str, Dict[str, Union[int, float, str]]] = {
            'session': {
                'cpu': {},
                'memory': {}
            }
        }

        for metrics_set in system_metrics_sets:

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                if metrics_sets.get(stage_name) is None:
                        metrics_sets[stage_name] = {
                            'cpu': {},
                            'memory': {}
                        }

                for monitor_name, monitor_metrics in stage_cpu_metrics.items():
                    metrics_sets[stage_name]['cpu'][monitor_name] = monitor_metrics.record

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_name, monitor_metrics in stage_memory_metrics.items():
                    metrics_sets[stage_name]['memory'][monitor_name] = monitor_metrics.record

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets[stage_name]['mb_per_vu'] = stage_mb_per_vu_metrics.record
                
            for monitor_name, monitor_metrics in metrics_set.session_cpu_metrics.items():
                metrics_sets['session']['cpu'][monitor_name] = monitor_metrics.record
                
            for monitor_name, monitor_metrics in metrics_set.session_memory_metrics.items():
                metrics_sets['session']['memory'][monitor_name] = monitor_metrics.record

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                json.dump,
                metrics_sets, 
                self.system_metrics_file, 
                indent=4
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Session System Metrics to file - {self.system_metrics_filepath}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):
        pass

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        if self.streams_file is None:
            self.streams_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.streams_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.streams_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Streams to file - {self.streams_filepath}')

        streams_data ={
            stream_name: stream_set.grouped for stream_name, stream_set in stream_metrics.items()
        }

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                json.dump,
                streams_data, 
                self.streams_file, 
                indent=4
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Streams to file - {self.streams_filepath}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Experiments to file - {self.experiments_filepath}')

        if self.experiments_file is None:
            self.experiments_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.experiments_filepath,
                    'w'
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.experiments_file
                    )
                )

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                json.dump,
                {
                    'experiments': experiment_metrics.experiments,
                    'variants': experiment_metrics.variants,
                    'mutations': experiment_metrics.mutations
                }, 
                self.experiments_file, 
                indent=4
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Experiments to file - {self.experiments_filepath}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):
        pass

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):
        pass

    async def submit_events(self, events: List[BaseProcessedResult]):

        if self.events_file is None:
            self.events_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.events_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.events_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Events to file - {self.events_filepath}')

        event_records = {
            event.event_id: event.record for event in events
        }

        if self.write_mode == 'a+':
            try:
                existing_data = await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        json.load,
                        self.events_file
                    )
                )
            
            except Exception:
                existing_data = {}

            event_records.update(existing_data)


        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                json.dump,
                event_records, 
                self.events_file, 
                indent=4
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Events to file - {self.events_filepath}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping Shared Metrics')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Metrics to file - {self.metrics_filepath}')

        records = {}
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            groups = {}
            for group_name, group in metrics_set.groups.items():
                groups[group_name] = group.record

            groups['custom'] = {
                metric.metric_shortname: metric.metric_value for metric in metrics_set.custom_metrics.values()
            }

            records[metrics_set.name] = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'errors': metrics_set.errors,
                **metrics_set.common_stats,
                'groups': groups
            }

        if self.metrics_file is None:
            self.metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.metrics_filepath,
                    self.write_mode
                )
            )

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                json.dump,
                records, 
                self.metrics_file, 
                indent=4
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Metrics to file - {self.metrics_filepath}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping Custom Metrics')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping Error Metrics')

    async def close(self):
        
        if self.events_file:
            await self._loop.run_in_executor(
                self._executor,
                self.events_file.close
            )

        if self.experiments_file:
            await self._loop.run_in_executor(
                self._executor,
                self.experiments_file.close
            )

        if self.metrics_file:
            await self._loop.run_in_executor(
                self._executor,
                self.metrics_file.close
            )  

        if self.streams_file:      
            await self._loop.run_in_executor(
                self._executor,
                self.streams_file.close
            )  

        if self.system_metrics_file:      
            await self._loop.run_in_executor(
                self._executor,
                self.system_metrics_file.close
            )  
            
        self._executor.shutdown(wait=False, cancel_futures=True)

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


import socket
import sys

def bind_tcp_socket(
    host: str,
    port: int
) -> socket.socket:

    family = socket.AF_INET

    if host and ":" in host:
        family = socket.AF_INET6

    sock = socket.socket(family, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    try:

        sock.bind((host, port))

    except OSError :
        sys.exit(1)
    
    sock.setblocking(False)
    sock.set_inheritable(True)

    return sock


def bind_udp_socket(
    host: str,
    port: int
) -> socket.socket:

    sock = socket.socket(
        socket.AF_INET, 
        socket.SOCK_DGRAM, 
        socket.IPPROTO_UDP
    )
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    try:

        sock.bind((host, port))

    except OSError :
        sys.exit(1)
    
    sock.setblocking(False)
    sock.set_inheritable(True)

    return sock


from .deform_header import DeformHeader
from .inject_header import InjectHeader
from .inject_junk_data import InjectJunkData
from .inject_ping import InjectPing
from .smuggle_request import SmuggleRequest


import json
import uuid
from datetime import datetime
from hedra.core.engines.types.playwright.command import PlaywrightCommand
from hedra.core.engines.types.common.types import RequestTypes
from typing import (
    Dict, 
    Any, 
    List, 
    Type,
    Union,
    Callable
)


try:
    from cassandra.cqlengine import columns
    from cassandra.cqlengine.models import Model
    has_connector = True

except Exception:
    columns = object
    Model = None
    has_connector = False


class CassandraPlaywrightActionSchema:

    def __init__(self, table_name: str) -> None:
        self.action_columns = {
            'id': columns.UUID(primary_key=True, default=uuid.uuid4),
            'name': columns.Text(min_length=1, index=True),
            'command': columns.Text(min_length=1),
            'selector': columns.Text(min_length=1),
            'attribute': columns.Text(min_length=1),
            'x_coordinate': columns.Text(min_length=1),
            'y_coordinate': columns.Text(min_length=1),
            'frame': columns.Integer(default=0),
            'location': columns.Text(),
            'headers': columns.Text(),
            'key': columns.Text(),
            'text': columns.Text(),
            'expression': columns.Text(),
            'args': columns.List(
                columns.Map(
                    columns.Text(),
                    columns.Text()
                )
            ),
            'filepath': columns.Text(),
            'file': columns.Text(),
            'path': columns.Text(),
            'option': columns.Map(
                columns.Text(),
                columns.Text()
            ),
            'by_label': columns.Boolean(default=False),
            'by_value': columns.Boolean(default=False),
            'event': columns.Text(),
            'is_checked': columns.Boolean(default=False),
            'timeout': columns.Float(),
            'extra': columns.Text(),
            'switch_by': columns.Text(),
            'user': columns.Text(),
            'tags': columns.List(
                columns.Map(
                    columns.Text(),
                    columns.Text()
                )
            ),
            'created_at': columns.DateTime(default=datetime.now)
        }

        self.actions_table = type(
            f'{table_name}_playwright',
            (Model, ), 
            self.action_columns
        )

        self.type = RequestTypes.PLAYWRIGHT

        self._types_map: Dict[type, str] = {
            str: 'string',
            int: 'integer',
            float: 'float',
            bool: 'boolean',
            bytes: 'bytes',
            dict: 'dictionary',
            list: 'list'
        }

        self._serialize_map: Dict[
            str, 
            Callable[
                [Type[
                    Union[
                        str, 
                        bytes, 
                        int, 
                        float, 
                        list, 
                        dict
                    ]
                ]],
                str
            ]
        ] = {
            str: str,
            bytes: str,
            int: str,
            float: str,
            bool: str,
            dict: json.dumps,
            list: json.dumps
        }
    

        self._deserialize_map: Dict[
            str, 
            Callable[
                [str],
                Type[
                    Union[
                        str, 
                        bytes, 
                        int, 
                        float, 
                        list, 
                        dict
                    ]
                ]
            ]
        ] = {
            'string': str,
            'integer': int,
            'float': float,
            'bool': lambda boolean: boolean == "True",
            'bytes': bytes,
            'dictionary': json.loads,
            'list': json.loads
        }
    
    def to_schema_record(
        self,
        action: PlaywrightCommand
    ) -> Dict[str, Any]:
        
        action_option: Union[Dict[str, str], None] = None
        if action.input.option:
            action_option = {
                    'data': self._serialize_map.get(type(
                        action.input.option
                    ))(action.input.option),
                    'datatype': self._types_map.get(type(
                        action.input.option
                    ))
            }


        return {
            'id': uuid.UUID(action.action_id),
            'name': action.name,
            'url': action.url,
            'selector': action.page.selector,
            'attribute': action.page.attribute,
            'x_coordinate': action.page.x_coordinate,
            'y_coordinate': action.page.y_coordinate,
            'frame': action.page.frame,
            'location': action.url.location,
            'headers': json.dumps(action.url.headers),
            'key': action.input.key,
            'text': action.input.text,
            'expression': action.input.expression,
            'args': [
                {
                    'data': self._serialize_map.get(type(
                        arg
                    ))(arg),
                    'datatype': self._types_map.get(type(
                        arg
                    ))
                } for arg in action.input.args
            ],
            'filepath': action.input.filepath,
            'file': action.input.file,
            'path': action.input.path,
            'option': action_option,
            'by_label': action.input.by_label,
            'by_value': action.input.by_value,
            'event': action.options.event,
            'is_checked': action.options.is_checked,
            'timeout': action.options.timeout,
            'extra': json.dumps(action.options.extra),
            'switch_by': action.options.switch_by,
            'user': action.metadata.user,
            'tags': [
                {
                    'name': tag.get('name'),
                    'value': str(tag.get('value')),
                    'datatype': self._types_map.get(type(
                        tag.get('value')
                    ))
                } for tag in action.metadata.tags
            ]
        }
    
    def from_schema_record(
        self,
        action: Dict[str, Any]
    ) -> Dict[str, Any]:
        
        action_tags: List[Dict[str, Any]] = action.get('tags', [])
        action_args: List[Dict[str, Any]] = action.get('args', [])
        action_option: Dict[str, Any] = action.get('option')

        option: Union[
            str, 
            bytes, 
            int, 
            float, 
            bool, 
            list, 
            dict, 
            None
        ] = None

        if action_option:
            option = self._deserialize_map.get(
                action_option.get('datatype')
            )(
                action_option.get('data')
            )

        return {
            'id': uuid.UUID(action.get('action_id')),
            'name': action.get('name'),
            'url': action.get('url'),
            'selector': action.get('selector'),
            'attribute': action.get('attribute'),
            'x_coordinate': action.get('x_coordinate'),
            'y_coordinate': action.get('y_coordinate'),
            'frame': action.get('frame'),
            'location': action.get('location'),
            'headers': json.loads(
                action.get('headers', {})    
            ),
            'key': action.get('key'),
            'text': action.get('text'),
            'expression': action.get('expression'),
            'args': [
                self._deserialize_map.get(
                    arg_config.get('datatype')
                )(
                    arg_config.get('data')
                ) for arg_config in action_args
            ],
            'filepath': action.get('filepath'),
            'file': action.get('file'),
            'path': action.get('path'),
            'option': option,
            'by_label': action.get('by_label'),
            'by_value': action.get('by_Value'),
            'event': action.get('event'),
            'is_checked': action.get('is_checked'),
            'timeout': action.get('timeout'),
            'extra': json.loads(
                action.get('extra', {})
            ),
            'switch_by': action.get('switch_by'),
            'user': action.get('user'),
            'tags': action_tags
        }



import re
from datetime import timedelta

class TimeParser:

    def __init__(self, time_amount: str) -> None:
        self.UNITS = {'s':'seconds', 'm':'minutes', 'h':'hours', 'd':'days', 'w':'weeks'}
        self.time = int(
            timedelta(
                **{
                    self.UNITS.get(
                        m.group(
                            'unit'
                        ).lower(), 
                        'seconds'
                    ): float(m.group('val')
                )
                    for m in re.finditer(
                        r'(?P<val>\d+(\.\d+)?)(?P<unit>[smhdw]?)', 
                        time_amount, 
                        flags=re.I
                    )
                }
            ).total_seconds()
        )



from typing import Tuple, Any
from types import SimpleNamespace
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.versioning.flags.types.unstable.flag import unstable_threadsafe
from .mutation_type import MutationType
from .validator import MutationValidator



class Mutation:

    def __init__(
        self,
        name: str,
        chance: float,
        mutation_type: MutationType,
        *targets: Tuple[str],
    ) -> None:
        validated_mutation = MutationValidator(
            name=name,
            chance=chance,
            targets=targets,
            mutation_type=mutation_type
        )

        self.name = validated_mutation.name
        self.chance = validated_mutation.chance
        self.targets = list(validated_mutation.targets)
        self.stage: Any = SimpleNamespace(
            context=SimpleContext()
        )

        self.mutation_type = validated_mutation.mutation_type

        unstable_threadsafe(Mutation)

    async def mutate(self, action: BaseAction=None):
        raise NotImplementedError(
            'Err. - mutate() is an abstract method in the base Mutation class.'
        )
    
    def copy(self):
        raise NotImplementedError(
            'Err. - copy() is an abstract method in the base Mutation class.'
        )


import inspect
from typing import Dict
from hedra.versioning.flags.types.unsafe.feature import UnsafeFeature
from hedra.versioning.flags.types.unstable.feature import UnstableFeature
from .active import active_flags
from .feature import Flag
from .flag_type import FlagTypes


class FlagRegistry:
    all: Dict[str, Flag] = {}
    module_paths: Dict[str, str] = {}

    def __init__(self, flag_type) -> None:
        self.flag_type = flag_type
        self.flag_types = {
            FlagTypes.UNSAFE_FEATURE: lambda *args, **kwargs: UnsafeFeature(
                *args, 
                **kwargs
            ),
            FlagTypes.UNSTABLE_FEATURE: lambda *args, **kwargs: UnstableFeature(
                *args, 
                **kwargs
            )
        }

    def __call__(self, flag):

        self.module_paths[flag.__name__] = flag.__module__

        def wrap_feature(feature):
            
            feature_name = feature.__name__

            flagged_feature: Flag = self.flag_types[self.flag_type]

                    
            self.all[feature_name] = flagged_feature(
                feature_name,
                feature
            )

            def wrapped_method(*args, **kwargs):

                selected_feature = self.all.get(feature_name)


                if active_flags.get(selected_feature.type):
                    selected_feature.enabled = True

                else:
                    raise selected_feature.exception(feature_name)

                return selected_feature.feature(*args, **kwargs)
            
            return wrapped_method

        return wrap_feature


def makeRegistrar():
    return FlagRegistry


flag_registrar = makeRegistrar()


from typing import Generic, TypeVar, List, Any, Dict
from hedra.core.hooks.types.condition.decorator import condition
from hedra.core.hooks.types.context.decorator import context
from hedra.core.hooks.types.event.decorator import event
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.plugins.types.plugin_types import PluginType
from hedra.reporting import Reporter
from hedra.reporting.experiment.experiment_metrics_set import ExperimentMetricsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.system.system_metrics_group import SystemMetricsGroup
from hedra.reporting.system.system_metrics_set_types import SystemMetricGroupType
from hedra.reporting.system import SystemMetricsSet
from typing import Optional, Union


T = TypeVar('T')


class Submit(Stage, Generic[T]):
    stage_type=StageTypes.SUBMIT
    stream: bool = False
    config: T= None
    priority: Optional[str]=None
    retries: int=0
    
    def __init__(self) -> None:
        super().__init__()
        self.accepted_hook_types = [ 
            HookType.CONDITION,
            HookType.CONTEXT,
            HookType.EVENT, 
            HookType.TRANSFORM
        ]

        self.source_internal_events = [
            'collect_process_results_and_metrics'
        ]

        self.internal_events = [
            'collect_process_results_and_metrics',
            'collect_reporter_plugins',
            'initialize_reporter',
            'check_for_events',
            'submit_processed_results',
            'submit_stage_metrics',
            'submit_main_metrics',
            'submit_error_metrics',
            'submit_custom_metrics',
            'complete_submit_session'
        ]

        self.stream = self.stream
        self.config = self.config
        self.priority = self.priority
        if self.priority is None:
            self.priority = 'auto'

        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

        self.stage_retries = self.retries

    @Internal()
    async def run(self):

        await self.setup_events()
        self.dispatcher.assemble_execution_graph()
        await self.dispatcher.dispatch_events(self.name)

    @context()
    async def collect_process_results_and_metrics(
        self,
        submit_stage_session_total: int = 0,
        submit_stage_events: List[T]=[],
        submit_stage_summary_metrics: List[MetricsSet]=[],
        submit_stage_experiment_metrics: List[ExperimentMetricsSet]=[],
        submit_stage_streamed_metrics: Dict[str, StageStreamsSet] = {},
        submit_stage_system_metrics: List[SystemMetricsSet]=[],
    ):
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Initializing results submission')

        self.context.ignore_serialization_filters = [
            'submit_stage_monitors',
            'session_stage_monitors'
        ]

        main_monitor_name = f'{self.name}.main'

        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        cpu_monitor.visibility_filters[main_monitor_name] = False
        memory_monitor.visibility_filters[main_monitor_name] = False

        cpu_monitor.stage_type = StageTypes.SUBMIT
        memory_monitor.stage_type = StageTypes.SUBMIT

        await cpu_monitor.start_background_monitor(main_monitor_name)
        await memory_monitor.start_background_monitor(main_monitor_name)

        return {
            'submit_stage_experiment_metrics': submit_stage_experiment_metrics,
            'submit_stage_session_total': submit_stage_session_total,
            'submit_stage_metrics': submit_stage_summary_metrics,
            'submit_stage_events': submit_stage_events,
            'submit_stage_streamed_metrics': submit_stage_streamed_metrics,
            'submit_stage_monitors': {
                'cpu': cpu_monitor,
                'memory': memory_monitor
            },
            'submit_stage_system_metrics': submit_stage_system_metrics
        }

    @event('collect_process_results_and_metrics')
    async def collect_reporter_plugins(self):

        reporter_plugins = self.plugins_by_type.get(PluginType.REPORTER)

        for plugin_name, plugin in reporter_plugins.items():
            Reporter.reporters[plugin_name] = plugin

            if isinstance(self.config, plugin.config):
                self.config.reporter_type = plugin_name
            
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Reporter plugin - {plugin_name}')

        return {
            'submit_stage_reporter_plugins': reporter_plugins
        }

    @event('collect_reporter_plugins')
    async def initialize_reporter(self):

        reporter = Reporter(self.config)
        reporter.graph_name = self.graph_name
        reporter.graph_id = self.graph_id
        reporter.stage_name = self.name
        reporter.stage_id = self.stage_id

        reporter_name = reporter.reporter_type_name

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Submitting results via - {reporter_name}:{reporter.reporter_id} - reporter')
        await self.logger.spinner.append_message(f'Submitting results via - {reporter_name} - reporter')

        await reporter.connect()


        return {
            'submit_stage_reporter': reporter,
            'submit_stage_reporter_name': reporter_name
        }
    
    @condition('initialize_reporter')
    async def check_for_streams(
        self,
        submit_stage_streamed_metrics: Dict[str, StageStreamsSet]={},
    ):
        return {
            'submit_stage_has_streams': len(submit_stage_streamed_metrics) > 0
        }
    
    @event('check_for_streams')
    async def submit_streamed_metrics(
        self,
        submit_stage_has_streams: bool=False,
        submit_stage_streamed_metrics: Dict[str, StageStreamsSet]={},
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None,
    ):
        if submit_stage_has_streams:
            submit_stage_streams_count = len(submit_stage_streamed_metrics)

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting - {submit_stage_streams_count} - Streams')
            await submit_stage_reporter.submit_streams(submit_stage_streamed_metrics)

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitted - {submit_stage_streams_count} - Streams')

    
    @condition('initialize_reporter')
    async def check_for_experiments(
        self,
        submit_stage_experiment_metrics: List[ExperimentMetricsSet]=[],

    ):
        return {
            'submit_stage_has_experiments': len(submit_stage_experiment_metrics) > 0
        }
    
    @event('check_for_experiments')
    async def submit_experiment_results(
        self,
        submit_stage_experiment_metrics: List[ExperimentMetricsSet]=[],
        submit_stage_has_experiments: bool = False,
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None,
    ):
        if submit_stage_has_experiments:
            submit_stage_experiments_count = len(submit_stage_experiment_metrics)

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting - {submit_stage_experiments_count} - Experiments')
            await submit_stage_reporter.submit_experiments(submit_stage_experiment_metrics)

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitted - {submit_stage_experiments_count} - Experiments')


    @condition('initialize_reporter')
    async def check_for_events(
        self,
        analyze_stage_events: List[Any]=[]
    ):
        return {
            'submit_stage_has_events': len(analyze_stage_events) > 0
        }

    @event('check_for_events')
    async def submit_processed_results(
        self,
        submit_stage_has_events: bool=False,
        submit_stage_events: List[Any]=[],
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None,
        submit_stage_session_total: int=0
    ):

        if submit_stage_has_events and self.stream is False:

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting - {submit_stage_session_total} - Events')
            await submit_stage_reporter.submit_events(submit_stage_events)

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitted - {submit_stage_session_total} - Events')

        return {}

    @event('initialize_reporter')
    async def submit_stage_metrics(
        self,
        submit_stage_metrics: List[Any]=[],
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None
    ):
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting Common Metrics')
        await submit_stage_reporter.submit_common(submit_stage_metrics)

        return {}

    @event('initialize_reporter')
    async def submit_main_metrics(
        self,
        submit_stage_metrics: List[Any]=[],
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None
    ):
        await submit_stage_reporter.submit_metrics(submit_stage_metrics)
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting Metrics')

        return {}

    @event('initialize_reporter')
    async def submit_error_metrics(
        self,
        submit_stage_metrics: List[Any]=[],
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None
    ):
        await submit_stage_reporter.submit_errors(submit_stage_metrics)
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting Error Metrics')

        return {}

    @event('initialize_reporter')
    async def submit_custom_metrics(
        self,
        submit_stage_metrics: List[Any]=[],
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None
    ):
        await submit_stage_reporter.submit_custom(submit_stage_metrics)
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting Custom Metrics')

        return {}
    
    @event(
        'submit_experiment_results',
        'submit_processed_results',
        'submit_stage_metrics',
        'submit_main_metrics',
        'submit_error_metrics',
        'submit_custom_metrics'
    )
    async def close_stage_system_monitors(
        self,
        submit_stage_system_metrics: List[SystemMetricsSet]=[],
        submit_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]]={}
    ):
        stage_cpu_monitor: CPUMonitor = submit_stage_monitors.get('cpu')
        stage_memory_monitor: MemoryMonitor = submit_stage_monitors.get('memory')

        main_monitor_name = f'{self.name}.main'
    
        await stage_cpu_monitor.stop_background_monitor(main_monitor_name)
        await stage_memory_monitor.stop_background_monitor(main_monitor_name)

        stage_cpu_monitor.close()
        stage_memory_monitor.close()

        stage_cpu_monitor.stage_metrics[main_monitor_name] = stage_cpu_monitor.collected[main_monitor_name]
        stage_memory_monitor.stage_metrics[main_monitor_name] = stage_memory_monitor.collected[main_monitor_name]  

        submit_stage_monitors = {
            self.name: {
                'cpu': stage_cpu_monitor,
                'memory': stage_memory_monitor
            }
        }    

        system_metrics = SystemMetricsSet(
            submit_stage_monitors,
            {}
        )

        system_metrics.generate_system_summaries()

        for metrics_set in submit_stage_system_metrics:

            metrics_set.metrics.update(submit_stage_monitors)
            metrics_set.cpu_metrics_by_stage[self.name] = stage_cpu_monitor
            metrics_set.memory_metrics_by_stage[self.name] = stage_memory_monitor

            metrics_set.cpu = SystemMetricsGroup(
                metrics_set.cpu_metrics_by_stage,
                SystemMetricGroupType.CPU 
            )

            metrics_set.memory = SystemMetricsGroup(
                metrics_set.memory_metrics_by_stage,
                SystemMetricGroupType.MEMORY 
            )

            metrics_set.generate_system_summaries()

        return {
            'submit_stage_system_metrics': submit_stage_system_metrics,
            'stage_system_metrics': system_metrics
        }

    @event('close_stage_system_monitors')
    async def submit_system_metrics(
        self,
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None,
        submit_stage_system_metrics: List[SystemMetricsSet]=[]
    ):
        
        await submit_stage_reporter.submit_system_metrics(submit_stage_system_metrics)
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Submitting System Metrics')

        return {}

    @context('submit_system_metrics')
    async def complete_submit_session(
        self,
        submit_stage_reporter: Reporter=None,
        submit_stage_reporter_name: str=None,
        submit_stage_session_total: int=0,
        submit_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]]={},
        stage_system_metrics: SystemMetricsSet=None
    ):

        await submit_stage_reporter.close()

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {submit_stage_reporter_name}:{submit_stage_reporter.reporter_id} - Completed Metrics submission')
        await self.logger.spinner.set_default_message(f'Successfully submitted the results for {submit_stage_session_total} actions via {submit_stage_reporter_name} reporter')

        return {
            'submit_stage_monitors': submit_stage_monitors,
            'stage_system_metrics': stage_system_metrics
        }
    
    @event('complete_submit_session')
    async def close_session(self):
        return {}


from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool,
    StrictInt,
    StrictFloat
)

from typing import Union


class ExecuteStageSummaryValidator(BaseModel):
    stage: StrictStr
    total_elapsed: Union[StrictInt, StrictFloat]
    total_results: Union[StrictInt, StrictFloat]
    stage_batch_size: StrictInt
    stage_optimized: StrictBool
    stage_persona_type: StrictStr
    stage_workers: StrictInt


from typing import Dict, List, Tuple, Union
from hedra.core.engines.types.common.decoder import Decoder
from hedra.core.engines.types.common.encoder import Encoder
from .base_event import BaseEvent


# Parsing headers mid load-test is *expensive* so we want to defer
# this work until later.
class DeferredHeaders(BaseEvent):
    event_type='DEFERRED_HEADERS'

    __slots__ = (
        'stream_id',
        'hpack_table',
        'raw_headers',
        'stream_ended',
        'end_stream',
        'priority',
        'encoding',
        'priority_updated'
    )

    def __init__(self, encoder: Encoder, frame, encoding: Union[str, None]) -> None:
        super().__init__()
        self.stream_id = frame.stream_id
        self.hpack_table = encoder.header_table
        self.raw_headers = frame.data
        self.stream_ended = None
        self.end_stream = 'END_STREAM' in frame.flags
        self.priority = 'PRIORITY' in frame.flags
        self.encoding = encoding
        self.priority_updated = None

    def parse(self) -> Tuple[int, Dict[bytes, bytes]]:
        decoder = Decoder()
        decoder.header_table = self.hpack_table
        decoder.header_table_size = self.hpack_table.maxsize
        headers: List[Tuple[bytes, bytes]] = {}

        try:
            headers = decoder.decode(self.raw_headers, raw=True)

        except Exception:
            return 400, {}

        status_code = None
        headers_dict: Dict[bytes, bytes] = {}
        for k, v in headers:
            if k == b":status":
                status_code = int(v.decode("ascii", errors="ignore"))
            elif k.startswith(b":"):
                headers_dict[k.strip(b':')] = v
            else:
                headers_dict[k] = v
        
        return status_code, headers_dict

    


import asyncio
import functools
import datetime
import json
import uuid
import psutil
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from concurrent.futures import ThreadPoolExecutor
from .cloudwatch_config import CloudwatchConfig

try:
    import boto3
    has_connector = True

except Exception:
    boto3 = None
    has_connector = False


class Cloudwatch:

    def __init__(self, config: CloudwatchConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name
        self.iam_role_arn = config.iam_role_arn
        self.schedule_rate = config.schedule_rate

        self.cloudwatch_targets = config.cloudwatch_targets
        self.aws_resource_arns = config.aws_resource_arns
        self.submit_timeout = config.submit_timeout

        self.events_rule_name = config.events_rule
        self.metrics_rule_name = config.metrics_rule
        self.streams_rule_name = config.streams_rule

        self.shared_metrics_rule_name = f'{config.metrics_rule}_shared'
        self.errors_rule_name = f'{config.metrics_rule}_errors'

        self.experiments_rule_name = config.experiments_rule
        self.variants_rule_name = f'{config.experiments_rule}_variants'
        self.mutations_rule_name = f'{config.experiments_rule}_mutations'

        self.session_system_metrics_rule_name = f'{config.system_metrics_rule}_session'
        self.stage_system_metrics_rule_name = f'{config.system_metrics_rule}_stage'

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self.experiments_rule = None

        self.client = None
        self._loop = asyncio.get_event_loop()

    async def connect(self):
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating AWS Cloudwatch client for region - {self.region_name}')

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'events',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created AWS Cloudwatch client for region - {self.region_name}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Cloudwatch rule - {self.session_system_metrics_rule_name}')

        rows: List[SessionMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')

            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                rows.append(monitor_metrics.record)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                rows.append(monitor_metrics.record)

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=[
                        {
                            'Time': datetime.datetime.now(),
                            'Detail': json.dumps(metrics_set.record),
                            'DetailType': self.session_system_metrics_rule_name,
                            'Resources': self.aws_resource_arns,
                            'Source': self.session_system_metrics_rule_name
                        } for metrics_set in rows
                    ]
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to Cloudwatch rule - {self.session_system_metrics_rule_name}')
    
    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Cloudwatch rule - {self.stage_system_metrics_rule_name}')

        rows: List[SystemMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    rows.append(monitor_metrics.record)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    rows.append(monitor_metrics.record)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    rows.append(stage_mb_per_vu_metrics.record)

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=[
                        {
                            'Time': datetime.datetime.now(),
                            'Detail': json.dumps(metrics_set.record),
                            'DetailType': self.stage_system_metrics_rule_name,
                            'Resources': self.aws_resource_arns,
                            'Source': self.stage_system_metrics_rule_name
                        } for metrics_set in rows
                    ]
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to Cloudwatch rule - {self.stage_system_metrics_rule_name}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Cloudwatch rule - {self.streams_rule_name}')

        streams = []
        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream Group - {group_name}:{stream.stream_set_id}')
                
                streams.append({
                    'Time': datetime.datetime.now(),
                    'Detail': json.dumps({
                        **group,
                        'group': group_name,
                        'stage': stage_name,
                        'name': f'{stage_name}_streams'
                    }),
                    'DetailType': self.streams_rule_name,
                    'Resources': self.aws_resource_arns,
                    'Source': self.streams_rule_name
                })

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=streams
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to Cloudwatch rule - {self.streams_rule_name}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Cloudwatch rule - {self.experiments_rule_name}')

        cloudwatch_events = [
            {
                'Time': datetime.datetime.now(),
                'Detail': json.dumps(experiment),
                'DetailType': self.experiments_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.experiments_rule_name
            } for experiment in experiment_metrics.experiments
        ]

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=cloudwatch_events
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Cloudwatch rule - {self.experiments_rule_name}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Cloudwatch rule - {self.variants_rule_name}')

        cloudwatch_events = [
            {
                'Time': datetime.datetime.now(),
                'Detail': json.dumps(variant),
                'DetailType': self.variants_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.variants_rule_name
            } for variant in experiment_metrics.variants
        ]

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=cloudwatch_events
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Cloudwatch rule - {self.variants_rule_name}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Cloudwatch rule - {self.mutations_rule_name}')

        cloudwatch_events = [
            {
                'Time': datetime.datetime.now(),
                'Detail': json.dumps(mutation),
                'DetailType': self.mutations_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.mutations_rule_name
            } for mutation in experiment_metrics.mutations
        ]

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=cloudwatch_events
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Cloudwatch rule - {self.mutations_rule_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Cloudwatch rule - {self.events_rule_name}')

        cloudwatch_events = [
            {
                'Time': datetime.datetime.now(),
                'Detail': json.dumps(event.record),
                'DetailType': self.events_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.events_rule_name
            } for event in events
        ]

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=cloudwatch_events
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Cloudwatch rule - {self.events_rule_name}')
    
    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Cloudwatch rule - {self.metrics_rule_name}')

        common_metrics = [
            {
                'Time': datetime.datetime.now(),
                'Detail': json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                }),
                'DetailType': self.shared_metrics_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.shared_metrics_rule_name
            } for metrics_set in metrics_sets
        ]

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=common_metrics
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Cloudwatch rule - {self.metrics_rule_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Cloudwatch rule - {self.metrics_rule_name}')

        metrics = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')
                
                metrics.append({
                    'Time': datetime.datetime.now(),
                    'Detail': json.dumps({
                        **group.record,
                        'group': group_name
                    }),
                    'DetailType': self.metrics_rule_name,
                    'Resources': self.aws_resource_arns,
                    'Source': self.metrics_rule_name
                })

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=metrics
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Cloudwatch rule - {self.metrics_rule_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Cloudwatch rule - {self.metrics_rule_name}')

        custom_metrics = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - Custom')

            custom_metrics.append({
                'Time': datetime.datetime.now(),
                'Detail': json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'custom',
                    **{
                        custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                    }
                }),
                'DetailType': self.metrics_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.metrics_rule_name
            })

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=custom_metrics
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Cloudwatch rule - {self.metrics_rule_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Cloudwatch rule - {self.errors_rule_name}')

        cloudwatch_errors = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                cloudwatch_errors.append({
                    'Time': datetime.datetime.now(),
                    'Detail': json.dumps(error),
                    'DetailType': self.errors_rule_name,
                    'Resources': self.aws_resource_arns,
                    'Source': self.errors_rule_name
                })

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=cloudwatch_errors
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Cloudwatch rule - {self.errors_rule_name}')


    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')



from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import BaseModel, StrictStr


class AWSLambdaConnectorConfig(BaseModel):
    aws_access_key_id: StrictStr
    aws_secret_access_key: StrictStr
    region_name: StrictStr
    lambda_name: StrictStr='actions'
    connector_type: ConnectorType=ConnectorType.AWSLambda


from .base_event import BaseEvent


class SettingsAcknowledged(BaseEvent):
    event_type='SETTINGS_ACKNOWLEDGED'
    __slots__ = ('changed_settings')
    """
    The SettingsAcknowledged event is fired whenever a settings ACK is received
    from the remote peer. The event carries on it the settings that were
    acknowedged, in the same format as
    :class:`h2.events.RemoteSettingsChanged`.
    """
    def __init__(self):
        #: A dictionary of setting byte to
        #: :class:`ChangedSetting <h2.settings.ChangedSetting>`, representing
        #: the changed settings.
        self.changed_settings = {}

    def __repr__(self):
        return "<SettingsAcknowledged changed_settings:{%s}>" % (
            ", ".join(repr(cs) for cs in self.changed_settings.values()),
        )




import asyncio
import time
import inspect
from typing import Dict, Generic, TypeVar
from hedra.plugins.types.common.plugin_hook import PluginHook
from hedra.plugins.types.common.types import PluginHooks
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.common.registrar import plugin_registrar
from hedra.core.personas.types.default_persona.default_persona import (
    DefaultPersona,
    cancel_pending
)
from hedra.core.engines.client.config import Config


T = TypeVar('T')


class PersonaPlugin(DefaultPersona, Generic[T], Plugin):
    type=PluginType.PERSONA
    name: str=None
    
    def __init__(self, config: Config):
        super().__init__(config)

        self.hooks: Dict[PluginHooks, PluginHook] = {}
        self.config = config

        methods = inspect.getmembers(self, predicate=inspect.ismethod) 
        for _, method in methods:

                method_name = method.__qualname__
                hook: PluginHook = plugin_registrar.all.get(method_name)
                
                if hook:
                    hook.call = hook.call.__get__(self, self.__class__)
                    setattr(self, hook.shortname, hook.call)

                    self.hooks[hook.hook_type] = hook

        self.name = self.name

    async def execute(self):

        setup_hook = self.hooks.get(PluginHooks.ON_PERSONA_SETUP)

        if setup_hook:
            await setup_hook.call()

        hooks = self._hooks
        loop = asyncio.get_running_loop()

        generate_hook = self.hooks.get(PluginHooks.ON_PERSONA_GENERATE)
        shutdown_hook = self.hooks.get(PluginHooks.ON_PERSONA_SHUTDOWN)

        await self.start_updates()

        self.start = time.monotonic()
        completed, pending = await asyncio.wait([
            loop.create_task(
                hooks[action_idx].session.execute_prepared_request(
                    hooks[action_idx].action
                )
            ) async for action_idx in generate_hook.call()
        ], timeout=self.graceful_stop)

        self.end = time.monotonic()
        self.pending_actions = len(pending)

        results = await asyncio.gather(*completed)
        
        await self.stop_updates()

        await asyncio.gather(*[
            asyncio.create_task(
                cancel_pending(pend)
            ) for pend in pending
        ])

        for hook in hooks:
            await hook.session.close()

        if shutdown_hook:
            await shutdown_hook.call()
        
        self.total_actions = len(set(results))
        self.total_elapsed = self.end - self.start
        self.optimized_params = None

        return results



from pydantic import (
    BaseModel,
    StrictStr
)
from hedra.data.connectors.common.connector_type import ConnectorType


class S3ConnectorConfig(BaseModel):
    aws_access_key_id: StrictStr
    aws_secret_access_key: StrictStr
    region_name: StrictStr
    bucket_name: StrictStr
    connector_type: ConnectorType=ConnectorType.S3


from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictInt
)
from typing import Optional, List


class DeformHeaderValidator(BaseModel):
    header_name: StrictStr
    deformation_length: StrictInt
    character_pool: Optional[List[StrictStr]]
    header_value: Optional[StrictStr]


from __future__ import annotations
import asyncio
import inspect
import traceback
from collections import defaultdict
from typing import Dict, List, Any, Union
from hedra.core.graphs.transitions.common.base_edge import BaseEdge
from hedra.core.hooks.types.base.event_types import EventType
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.setup.setup import Setup
from hedra.core.graphs.stages.execute import Execute
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.reporting.system.system_metrics_set_types import MonitorGroup


class SetupEdge(BaseEdge[Setup]):

    def __init__(self, source: Setup, destination: BaseEdge[Stage]) -> None:
        super(
            SetupEdge,
            self
        ).__init__(
            source,
            destination
        )

        self.valid_states = [
            StageStates.INITIALIZED,
            StageStates.VALIDATED
        ]

        self.requires = [
            'execute_stage_streamed_analytics',
            'execute_stage_results',
            'session_stage_monitors'
        ]

        self.provides = [
            'setup_stage_experiment_config',
            'execute_stage_streamed_analytics',
            'execute_stage_setup_hooks',
            'execute_stage_setup_config',
            'execute_stage_setup_by',
            'setup_stage_ready_stages',
            'setup_stage_candidates',
            'setup_stage_configs',
            'execute_stage_results',
            'session_stage_monitors',
            'setup_stage_monitors'
        ]

        self.assigned_candidates = []
        self.assigned_optimize_candidates = []

    async def transition(self):
        
        try:
            self.source.state = StageStates.SETTING_UP

            setup_candidates = self.get_setup_candidates()
            optimize_candidates = self.get_optimize_candidates()

            self.source.generation_setup_candidates = len(setup_candidates)

            for setup_candidate in setup_candidates.values():
                if setup_candidate.context is None:
                    setup_candidate.context = SimpleContext()

            
            for optimize_candidate in optimize_candidates.values():
                if optimize_candidate.context is None:
                    optimize_candidate.context = SimpleContext()

            self.edge_data['setup_stage_target_stages'] = setup_candidates
            self.edge_data['setup_stage_target_config'] = self.source.config

            self.source.context.update(self.edge_data)

            for event in self.source.dispatcher.events_by_name.values():
                event.source.stage_instance = self.source
                self.source.context.update(self.edge_data)
                event.context.update(self.edge_data)
                
                if event.source.context:
                    event.source.context.update(self.edge_data)
            
            if self.timeout and self.skip_stage is False:
                await asyncio.wait_for(self.source.run(), timeout=self.timeout)
            elif self.skip_stage is False:
                await self.source.run()

            for provided in self.provides:
                self.edge_data[provided] = self.source.context[provided]

            self.edge_data['setup_stage_candidates'] = setup_candidates

            self._update(self.destination)

            if self.destination.context is None:
                self.destination.context = SimpleContext()

            for execute_stage in setup_candidates.values():

                if execute_stage.name != self.destination.name:
                    execute_stage.state = StageStates.SETUP

                    if execute_stage.context is None:
                        execute_stage.context = SimpleContext()
                    
                    self._update(execute_stage)

            
            for optimize_stage in optimize_candidates.values():

                if optimize_stage.name != self.destination.name:
                    optimize_stage.state = StageStates.SETUP

                    if optimize_stage.context is None:
                        optimize_stage.context = SimpleContext()

                    self._update(optimize_stage)
            
            self.visited.append(self.source.name)

        except Exception as edge_exception:

            self.exception = edge_exception

        return None, self.destination.stage_type

    def _update(self, destination: Stage):

        for edge_name in self.history:

            history = self.history[edge_name]

            if self.next_history.get(edge_name) is None:
                self.next_history[edge_name] = {}

            self.next_history[edge_name].update(history)


        if self.next_history.get((self.source.name, destination.name)) is None:
            self.next_history[(self.source.name, destination.name)] = {}

        if self.skip_stage is False:
            setup_stage_configs = self.edge_data.get('setup_stage_configs', {})
            ready_stages = self.edge_data.get('setup_stage_ready_stages', {})
            setup_candidates = self.edge_data.get('setup_stage_candidates', {})
            setup_config = self.edge_data.get('execute_stage_setup_config')
            setup_stage_experiment_config: Dict[str, Union[str, int, List[float]]] = self.edge_data.get('setup_stage_experiment_config', {})
            execute_stage_setup_hooks = []
            setup_execute_stage: Execute = ready_stages.get(self.source.name)

            if setup_execute_stage:
                execute_stage_setup_hooks = setup_execute_stage.context['execute_stage_setup_hooks']

            self.stages_by_type[StageTypes.EXECUTE].update(ready_stages)

            session_stage_monitors = self.edge_data['session_stage_monitors']

            session_stage_monitors.update(
                self.edge_data['setup_stage_monitors']
            )

            self.next_history[(self.source.name, destination.name)].update({
                'setup_stage_experiment_config': setup_stage_experiment_config,
                'setup_stage_configs': setup_stage_configs,
                'execute_stage_setup_hooks': execute_stage_setup_hooks,
                'setup_stage_ready_stages': ready_stages,
                'setup_stage_candidates': list(setup_candidates.keys()),
                'execute_stage_setup_config': setup_config,
                'execute_stage_setup_by': self.source.name,
                'session_stage_monitors': session_stage_monitors
            })
            

    def split(self, edges: List[SetupEdge]) -> None:

        setup_stage_config: Dict[str, Any] = self.source.to_copy_dict()

        setup_stage_copy = type(self.source.name, (Setup, ), {})()
        
        for copied_attribute_name, copied_attribute_value in setup_stage_config.items():
            if inspect.ismethod(copied_attribute_value) is False:
                setattr(setup_stage_copy, copied_attribute_name, copied_attribute_value)

        user_hooks: Dict[str, Dict[str, Hook]] = defaultdict(dict)
        for hooks in registrar.all.values():
            for hook in hooks:
                if hasattr(self.source, hook.shortname) and not hasattr(Execute, hook.shortname):
                    user_hooks[self.source.name][hook.shortname] = hook._call

        setup_stage_copy.dispatcher = self.source.dispatcher.copy()

        minimum_edge_idx = min([edge.transition_idx for edge in edges])

        setup_stage_copy.context = SimpleContext()
        for event in setup_stage_copy.dispatcher.events_by_name.values():
            event.context = setup_stage_copy.context 
            event.source.stage_instance = setup_stage_copy
            event.source.stage_instance.context = setup_stage_copy.context
            event.source.context = setup_stage_copy.context

            if event.source.shortname in user_hooks[setup_stage_copy.name]:
                hook_call = user_hooks[setup_stage_copy.name].get(event.source.shortname)

                hook_call = hook_call.__get__(setup_stage_copy, setup_stage_copy.__class__)
                setattr(setup_stage_copy, event.source.shortname, hook_call)

                event.source._call = hook_call

            else:            
                event.source._call = getattr(setup_stage_copy, event.source.shortname)
                event.source._call = event.source._call.__get__(setup_stage_copy, setup_stage_copy.__class__)
                setattr(setup_stage_copy, event.source.shortname, event.source._call)
          
        self.source = setup_stage_copy

        if minimum_edge_idx < self.transition_idx:
            self.skip_stage = True

    def get_setup_candidates(self) -> Dict[str, Execute]:
        execute_stages: Dict[str, Execute] = self.stages_by_type.get(StageTypes.EXECUTE)
        setup_stages: Dict[str, Setup] = self.stages_by_type.get(StageTypes.SETUP)
        path_lengths: Dict[str, int] = self.path_lengths.get(self.source.name)

        all_paths = self.all_paths.get(self.source.name, [])

        execute_candidates: Dict[str, Stage] = {}

        for stage_name, stage in execute_stages.items():
            if stage_name in all_paths:
                execute_candidates[stage_name] = stage    

        selected_execute_candidates: Dict[str, Execute] = {}
        
        following_setup_stage_distances = [
            path_length for stage_name, path_length in path_lengths.items() if stage_name in setup_stages
        ]

        for stage_name in path_lengths.keys():
            stage_distance = path_lengths.get(stage_name)

            if stage_name in execute_candidates:

                if len(following_setup_stage_distances) > 0 and stage_distance <= min(following_setup_stage_distances):
                    selected_execute_candidates[stage_name] = execute_candidates.get(stage_name)

                elif len(following_setup_stage_distances) == 0:
                    selected_execute_candidates[stage_name] = execute_candidates.get(stage_name)

        for candidate in selected_execute_candidates.values():
            actions = [event for event in candidate.dispatcher.actions_and_tasks.values() if event.event_type == EventType.ACTION]
            tasks = [event for event in candidate.dispatcher.actions_and_tasks.values() if event.event_type == EventType.TASK]

            candidate.hooks[HookType.ACTION] = actions
            candidate.hooks[HookType.TASK] = tasks

        return selected_execute_candidates
    
    def get_optimize_candidates(self) -> Dict[str, Execute]:
        optimize_stages: Dict[str, Execute] = self.stages_by_type.get(StageTypes.OPTIMIZE, {})
        setup_stages: Dict[str, Setup] = self.stages_by_type.get(StageTypes.SETUP)
        path_lengths: Dict[str, int] = self.path_lengths.get(self.source.name)

        all_paths = self.all_paths.get(self.source.name, [])

        optimize_candidates: Dict[str, Stage] = {}

        for stage_name, stage in optimize_stages.items():
            if stage_name in all_paths:
                optimize_candidates[stage_name] = stage    

        selected_optimize_candidates: Dict[str, Execute] = {}
        
        following_setup_stage_distances = [
            path_length for stage_name, path_length in path_lengths.items() if stage_name in setup_stages or stage_name in optimize_stages
        ]

        for stage_name in path_lengths.keys():
            stage_distance = path_lengths.get(stage_name)

            if stage_name in optimize_candidates:

                if len(following_setup_stage_distances) > 0 and stage_distance <= min(following_setup_stage_distances):
                    selected_optimize_candidates[stage_name] = optimize_candidates.get(stage_name)

                elif len(following_setup_stage_distances) == 0:
                    selected_optimize_candidates[stage_name] = optimize_candidates.get(stage_name)

        return selected_optimize_candidates
    
    def setup(self) -> None:

        session_stage_monitors: MonitorGroup = {}

        for source_stage, destination_stage in self.history:
            
            previous_history: Dict[str, Any] = self.history[(source_stage, destination_stage)]
            
            stage_monitors = previous_history.get('session_stage_monitors')
            if stage_monitors:
                session_stage_monitors.update(stage_monitors)

        self.edge_data = {
            'session_stage_monitors': session_stage_monitors
        }





import asyncio
import dill
import functools
import statistics
import signal
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import Union, List, Dict, Any, Tuple
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.hooks.types.condition.decorator import condition
from hedra.core.hooks.types.context.decorator import context
from hedra.core.hooks.types.event.decorator import event  
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.event_types import EventType
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from hedra.logging import logging_manager
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.plugins.types.plugin_types import PluginType
from hedra.reporting.experiment.experiment_metrics_set import ExperimentMetricsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.metric.stage_metrics_summary import StageMetricsSummary
from hedra.reporting.metric.custom_metric import CustomMetric
from hedra.reporting.processed_result import results_types, ProcessedResultsGroup
from hedra.reporting.processed_result.types import (
    GraphQLProcessedResult,
    GraphQLHTTP2ProcessedResult,
    GRPCProcessedResult,
    HTTPProcessedResult,
    HTTP2ProcessedResult,
    PlaywrightProcessedResult,
    TaskProcessedResult,
    UDPProcessedResult,
    WebsocketProcessedResult,
)
from hedra.reporting.system import SystemMetricsSet
from hedra.reporting.system.system_metrics_set_types import MonitorGroup
from hedra.versioning.flags.types.base.active import active_flags
from hedra.versioning.flags.types.base.flag_type import FlagTypes
from typing import Optional
from .parallel import process_results_batch

dill.settings['byref'] = True


Events = Union[
    GraphQLProcessedResult, 
    GraphQLHTTP2ProcessedResult, 
    GRPCProcessedResult, 
    HTTPProcessedResult, 
    HTTP2ProcessedResult, 
    PlaywrightProcessedResult, 
    TaskProcessedResult,
    UDPProcessedResult,
    WebsocketProcessedResult
]

StageConfig = Tuple[str, int, List[Dict[str, List[BaseResult]]]]

RawResultsSet = Dict[str, ResultsSet]

RawResultsPairs = List[Tuple[Dict[str,  List[Tuple[str, Any]]]]]

ProcessedResults = Dict[str, Union[Dict[str, Union[int, float, int]], int]]

ProcessedMetricsSet = Dict[str, MetricsSet]

ProcessedStageMetricsSet = Dict[str, Union[int, float, Dict[str, ProcessedMetricsSet]]]

ProcessedResultsSet = List[Tuple[str, StageMetricsSummary]]

CustomMetricsSet = Dict[str, Dict[str, Dict[str, Union[int, float, Any]]]]

EventsSet = Dict[str, Dict[str, ProcessedResultsGroup]]

MonitorResults = Dict[str, List[Union[int, float]]]


def handle_loop_stop(
    signame,
    loop: asyncio.AbstractEventLoop,
    executor: ThreadPoolExecutor
):
    try:
        executor.shutdown(wait=False, cancel_futures=True)
        loop.close()

    except BrokenPipeError:
        pass

    except RuntimeError:
        pass


def deserialize_results(results: List[bytes]) -> List[BaseResult]:
    return [
        dill.loads(result) for result in results
    ]


class Analyze(Stage):
    stage_type=StageTypes.ANALYZE
    is_parallel=False
    handler=None
    priority: Optional[str]=None
    retries: int = 0

    def __init__(self) -> None:
        super().__init__()

        self.accepted_hook_types = [ 
            HookType.CONDITION,
            HookType.CONTEXT,
            HookType.CONDITION,
            HookType.EVENT, 
            HookType.METRIC,
            HookType.TRANSFORM, 
        ]

        self.stage_retries = self.retries
        self.requires_shutdown = True
        self.allow_parallel = True
        self.analysis_execution_time = 0
        self.analysis_execution_time_start = 0
        self._executor = ThreadPoolExecutor(max_workers=self.workers)
        self._loop: asyncio.AbstractEventLoop = None
        self.source_internal_events = [
            'initialize_raw_results'
        ]

        self.internal_events = [
            'initialize_results_analysis',
            'partition_results_batches',
            'get_custom_metric_hooks',
            'create_stage_batches',
            'assign_stage_batches',
            'analyze_stage_batches',
            'reduce_stage_contexts',
            'merge_events_groups',
            'calculate_custom_metrics',
            'generate_metrics_sets',
            'generate_summary',
            'complete'
        ]

        self.priority = self.priority
        if self.priority is None:
            self.priority = 'auto'

        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

    @Internal()
    async def run(self):  
        self.executor.batch_by_stages = True

        await self.setup_events()
        self.dispatcher.assemble_execution_graph()
        await self.dispatcher.dispatch_events(self.name)

    @context()
    async def initialize_results_analysis(
        self,
        analyze_stage_raw_results: RawResultsSet={},
        session_stage_monitors: MonitorGroup={},
        session_setup_stage_configs: Dict[str, Config]={}
    ):

        await self.logger.filesystem.aio.create_logfile('hedra.reporting.log')
        self.logger.filesystem.create_filelogger('hedra.reporting.log')
        
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting results analysis')

        self.analysis_execution_time_start = time.monotonic()

        engine_plugins = self.plugins_by_type.get(PluginType.ENGINE)
        for plugin_name, plugin in engine_plugins.items():
            results_types[plugin_name] = plugin.event

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Generated custom Event - {plugin.event.type} - for Reporter plugin - {plugin_name}')

        self.context.ignore_serialization_filters = [
            'analyze_stage_all_results',
            'analyze_stage_raw_results',
            'analyze_stage_target_stages',
            'analyze_stage_deserialized_results',
            'session_stage_monitors',
            'analyze_stage_monitors'
        ]

        all_results = list(analyze_stage_raw_results.items())

        total_group_results = 0
        for stage_results in analyze_stage_raw_results.values():
            total_group_results += stage_results.total_results
        
        return {
            'analyze_stage_raw_results': analyze_stage_raw_results,
            'analyze_stage_all_results': all_results,
            'analyze_stage_stages_count': len(analyze_stage_raw_results),
            'analyze_stage_total_group_results': total_group_results,
            'session_setup_stage_configs': session_setup_stage_configs,
            'session_stage_monitors': session_stage_monitors
        }
    
    @event('initialize_results_analysis')
    async def add_shutdown_handler(self):
        self._loop = asyncio.get_running_loop()

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._loop,
                    self.executor
                )
            )

    @condition('add_shutdown_handler')
    async def check_if_has_multiple_workers(self):
        return {
            'analyze_stage_has_multiple_workers': self.total_pool_cpus > 1
        }

    @event('check_if_has_multiple_workers')
    async def generate_deserialized_results(
        self,
        analyze_stage_raw_results: RawResultsSet={},
        analyze_stage_has_multiple_workers: bool=False
    ):
        deserialized_results = dict(analyze_stage_raw_results)
        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        main_monitor_name = f'{self.name}.main'

        await cpu_monitor.start_background_monitor(main_monitor_name)
        await memory_monitor.start_background_monitor(main_monitor_name)

        if analyze_stage_has_multiple_workers:
            for results_set_name, results_set in analyze_stage_raw_results.items():
                
                results_set_copy = results_set.copy()
                results_set_copy.results = await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        deserialize_results,
                        results_set.results
                    )
                )

                deserialized_results[results_set_name] = results_set_copy

        return {
            'analyze_stage_monitors': {
                'cpu': cpu_monitor,
                'memory': memory_monitor
            },
            'analyze_stage_raw_results': analyze_stage_raw_results,
            'analyze_stage_deserialized_results': deserialized_results
        }    


    @event('generate_deserialized_results')
    async def partition_results_batches(
        self,
        analyze_stage_raw_results: RawResultsSet={},
        analyze_stage_all_results: List[RawResultsPairs]=[]
    ):
        batches = self.executor.partion_stage_batches(analyze_stage_all_results)

        elapsed_times = []
        for stage_name, _, _ in batches:
            stage_results: ResultsSet = analyze_stage_raw_results.get(stage_name)
            elapsed_times.append(
                stage_results.total_elapsed
            )

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Paritioned {len(batches)} batches of results')

        return {
            'analyze_stage_batches': batches,
            'analyze_stage_elapsed_times': elapsed_times
        }
    
    @event('partition_results_batches')
    async def create_stage_batches(
        self,
        analyze_stage_raw_results: RawResultsSet=[],
        analyze_stage_batches: List[Tuple[str, Any, int]]=[],
        session_setup_stage_configs: Dict[str, Config]={}
    ):
        stage_total_times = {}
        stage_batch_sizes = {}
        stage_streamed_analytics: Dict[str, List[StreamAnalytics]] = {}
        analyze_stage_batch_configs = {}
        stage_personas = {}

        for stage_name, _, assigned_workers_count in analyze_stage_batches:
            
            stage_batches: List[List[Any]] = []

            stage_results = analyze_stage_raw_results.get(stage_name)
            results = stage_results.results
            stage_total_time = stage_results.total_elapsed
            
            stage_total_times[stage_name] = stage_total_time
            stage_streamed_analytics[stage_name] = stage_results.stage_streamed_analytics

            results_count = len(results)
            
            batch_size = int(results_count/assigned_workers_count)
            
            for worker_idx in range(assigned_workers_count):

                batch_marker = worker_idx * batch_size

                stage_batches.append(
                    results[batch_marker:batch_marker + batch_size]
                )

            if results_count%assigned_workers_count > 0:
                stage_batches[assigned_workers_count-1].extend(
                    results[assigned_workers_count * batch_size:]
                )

            analyze_stage_batch_configs[stage_name] = stage_batches
            stage_batch_sizes[stage_name] = stage_results.stage_batch_size
            stage_personas[stage_name] = stage_results.stage_persona_type

        return {
            'analyze_stage_target_stages': {},
            'analyze_stage_batch_configs': analyze_stage_batch_configs,
            'analyze_stage_total_times': stage_total_times,
            'analyze_stage_batch_sizes': stage_batch_sizes,
            'analyze_stage_personas': stage_personas,
            'analyze_stage_streamed_analytics': stage_streamed_analytics
        }

    @context('create_stage_batches')
    async def assign_stage_batches(
        self,
        analyze_stage_batches: List[Tuple[str, Any, int]]=[],
        analyze_stage_batch_configs: Dict[str, List[List[Any]]]=[],
        analyze_stage_metric_hook_names: List[str]=[],
        analyze_stage_has_multiple_workers: bool=False
    ):
        if analyze_stage_has_multiple_workers:
            stage_configs = []
            serializable_context = self.context.as_serializable()
            worker_idx = 0

            for stage_name, _, assigned_workers_count in analyze_stage_batches:

                stage_configs.append((
                    stage_name,
                    assigned_workers_count,
                    [
                        {
                            'graph_name': self.graph_name,
                            'graph_path': self.graph_path,
                            'graph_id': self.graph_id,
                            'enable_unstable_features': active_flags[FlagTypes.UNSTABLE_FEATURE],
                            'logfiles_directory': logging_manager.logfiles_directory,
                            'log_level': logging_manager.log_level_name,
                            'source_stage_name': self.name,
                            'source_stage_context': {
                                context_key: context_value for context_key, context_value in serializable_context
                            },
                            'source_stage_id': self.stage_id,
                            'analyze_stage_name': stage_name,
                            'analyze_stage_metric_hooks': list(analyze_stage_metric_hook_names),
                            'analyze_stage_batched_results': batch,
                            'worker_id': worker_idx + 1
                        } for batch in analyze_stage_batch_configs[stage_name]
                    ]
                ))
                
                worker_idx += 1
                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Assigned {assigned_workers_count} to process results from stage - {stage_name}')

            return {
                'analyze_stage_configs': stage_configs,
            }
    
    @condition('assign_stage_batches')
    async def check_if_multiple_stages(
        self,
        analyze_stage_configs: List[Tuple[str, Any, int]]=[],
    ):
       return {
           'multiple_stages_to_process': len(analyze_stage_configs) > 1
       } 

    @event('check_if_multiple_stages')
    async def execute_batched_analysis(
        self,
        analyze_stage_stages_count: int=0,
        analyze_stage_elapsed_times: List[float]=[],
        analyze_stage_total_group_results: int=0,
        analyze_stage_configs: List[StageConfig]=[],
        multiple_stages_to_process: bool=False,
        analyze_stage_deserialized_results: RawResultsSet = {},
        analyze_stage_has_multiple_workers: bool=False,
    ):

        if multiple_stages_to_process and analyze_stage_has_multiple_workers:
            await self.logger.spinner.append_message(
                f'Calculating results for - {analyze_stage_stages_count} - stages'
            )

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Processing results or - {analyze_stage_stages_count} - stages')
            
            median_execution_time = round(statistics.median(analyze_stage_elapsed_times))
            await self.logger.spinner.append_message(f'Calculating stats for - {analyze_stage_total_group_results} - actions executed over a median stage execution time of {median_execution_time} seconds')

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Calculating stats for - {analyze_stage_total_group_results} - actions over a median stage execution time of {median_execution_time} seconds')
            stage_batch_results = await self.executor.execute_batches(
                analyze_stage_configs,
                process_results_batch
            )

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Completed parital results aggregation for - {analyze_stage_stages_count} - stages')

            return {
                'analyze_stage_batch_results': stage_batch_results
            }
        
        else:
            

            batch_results = []
            events =  defaultdict(ProcessedResultsGroup)

            for results_set in analyze_stage_deserialized_results.values():

                for stage_result in results_set.results:
                    events[stage_result.name].add(
                            results_set.stage,
                            stage_result,
                        )
                    
                batch_results.append(
                    (results_set.stage, events)
                )
                
                for events_group in events.values():
                    events_group.calculate_stats()
                    
            return {
                'analyze_stage_batch_results': batch_results
            }
        
    @event('execute_batched_analysis')
    async def merge_events_groups(
        self,
        analyze_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]]={},
        analyze_stage_batch_results: List[Tuple[str, List[Dict[str, Any]]]]=[],
        multiple_stages_to_process: bool=False
    ):

        stage_events_set = {}

        stage_cpu_monitor: CPUMonitor = analyze_stage_monitors.get('cpu')
        stage_memory_monitor: MemoryMonitor = analyze_stage_monitors.get('memory')

        if multiple_stages_to_process:

            stage_cpu_monitor.stage_type = StageTypes.ANALYZE
            stage_memory_monitor.stage_type = StageTypes.ANALYZE

            for stage_name, stage_results in analyze_stage_batch_results:
                results = stage_results.pop()

                worker_id = results.get('worker_id')
                stage_events_set[stage_name] = results.get('events')

                monitors: Dict[str, MonitorResults] = results.get('monitoring', {})
                cpu_monitor = monitors.get('cpu', {})
                memory_monitor = monitors.get('memory', {})
                
                for monitor_name, collection_stats in cpu_monitor.items():
                    stage_cpu_monitor.worker_metrics[worker_id][monitor_name] = collection_stats
                    stage_cpu_monitor.collected[monitor_name].extend(collection_stats)

                for monitor_name, collection_stats in memory_monitor.items():
                    stage_memory_monitor.worker_metrics[worker_id][monitor_name] = collection_stats

            stage_cpu_monitor.aggregate_worker_stats()
            stage_memory_monitor.aggregate_worker_stats()

        else:

            for stage_name, stage_results in analyze_stage_batch_results:
                stage_events_set[stage_name] = stage_results

        main_monitor_name = f'{self.name}.main'

        await stage_cpu_monitor.stop_background_monitor(main_monitor_name)
        await stage_memory_monitor.stop_background_monitor(main_monitor_name)

        stage_cpu_monitor.stage_metrics[main_monitor_name] = stage_cpu_monitor.collected[main_monitor_name]
        stage_memory_monitor.stage_metrics[main_monitor_name] = stage_memory_monitor.collected[main_monitor_name]

        stage_cpu_monitor.close()
        stage_memory_monitor.close()

        return {
            'analyze_stage_events_set': stage_events_set,
            'analyze_stage_monitors': {
                'cpu': stage_cpu_monitor,
                'memory': stage_memory_monitor
            }
        }

    @event('merge_events_groups')
    async def calculate_custom_metrics(self):

        custom_metrics_set = defaultdict(dict)

        metrics = [
            metric_event for metric_event in self.dispatcher.events[EventType.METRIC]
        ]

        for metric in metrics:
            for context_key, context_value in metric.context:
                if isinstance(context_value, CustomMetric):
                    custom_metrics_set[context_value.metric_group][context_key] = context_value

        return {
            'analyze_stage_custom_metrics_set': custom_metrics_set
        }

    @event('calculate_custom_metrics')
    async def generate_metrics_sets(
        self,
        analyze_stage_custom_metrics_set: CustomMetricsSet={},
        analyze_stage_events_set: EventsSet={},
        analyze_stage_total_times: Dict[str, float]={},
        analyze_stage_personas: Dict[str, str]={},
        analyze_stage_batch_sizes: Dict[str, int]={},
        analyze_stage_streamed_analytics: Dict[str, List[StreamAnalytics]] = {}
    ):

        processed_results = []
        
        for stage_name, stage_events in analyze_stage_events_set.items():    

            stage_total_time = analyze_stage_total_times.get(stage_name)

            persona_type = analyze_stage_personas.get(stage_name)
            batch_size = analyze_stage_batch_sizes.get(stage_name)

            stage_metrics_summary = StageMetricsSummary(
                stage_name=stage_name,
                persona_type=persona_type,
                batch_size=batch_size,
                total_elapsed=stage_total_time,
                stage_streamed_analytics=analyze_stage_streamed_analytics.get(stage_name)
            )
            grouped_stats = {}

            for event_group_name, events_group in stage_events.items():  

                custom_metrics = analyze_stage_custom_metrics_set.get(event_group_name, {})
                
                events_group.calculate_quantiles()

                metric_data = {
                    'total': events_group.total,
                    'succeeded': events_group.succeeded,
                    'failed': events_group.failed,
                    'actions_per_second': round(
                        events_group.total/stage_total_time,
                        2
                    ),
                    'errors': list([
                        {
                            'message': error_message,
                            'count': error_count
                        } for error_message, error_count in events_group.errors.items()
                    ]),
                    'groups': events_group.groups,
                    'custom': custom_metrics
                }

                metric = MetricsSet(
                    event_group_name,
                    events_group.source,
                    stage_name,
                    metric_data,
                    events_group.tags
                )

                stage_metrics_summary.metrics_sets[event_group_name] = metric

                grouped_stats[event_group_name] = metric

                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Convererted stats for stage - {stage_name} to metrics set')
        
            stage_metrics_summary.calculate_action_and_task_metrics()
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Calculated results for - {stage_metrics_summary.stage_metrics.total} - actions from stage - {stage_name}')

            processed_results.append((
                stage_name,
                stage_metrics_summary
            ))

        return {
            'analyze_stage_processed_results': processed_results
        }
    
    @event('generate_metrics_sets')
    async def generate_experiment_metrics(
        self,
        analyze_stage_processed_results: ProcessedResultsSet=[],
        analyze_stage_raw_results: RawResultsSet=[],
    ):
        
        experiment_metrics_sets: Dict[str, ExperimentMetricsSet] = {}

        for stage_name, stage_metrics in analyze_stage_processed_results:

            
            stage_metrics_sets = stage_metrics.metrics_sets

            raw_results_set = analyze_stage_raw_results.get(stage_name)
            experiment = raw_results_set.experiment
            
            if experiment:
                variant = experiment.get('experiment_variant')
                variant_name = variant.get('variant_stage')
                mutations = variant.get('variant_mutations')

                experiment_name = experiment.get('experiment_name')


                experiment_metrics_set = experiment_metrics_sets.get(experiment_name)
                if experiment_metrics_set is None:
                    experiment_metrics_set = ExperimentMetricsSet()

                if experiment_metrics_set.experiment_name is None:
                    experiment_metrics_set.experiment_name = experiment.get('experiment_name')

                
                if experiment_metrics_set.randomized is None:
                    experiment_metrics_set.randomized = experiment.get('experiment_randomized')

                experiment_metrics_set.participants.append(stage_name)

                variant['stage_batch_size'] = raw_results_set.stage_batch_size
                variant['stage_optimized'] = raw_results_set.stage_optimized
                variant['stage_persona_type'] = raw_results_set.stage_persona_type
                variant['stage_workers'] = raw_results_set.stage_workers

                experiment_metrics_set.variants[variant_name] = variant
                experiment_metrics_set.mutations[variant_name] = mutations
                experiment_metrics_set.metrics[variant_name] = stage_metrics_sets

                experiment_metrics_sets[experiment_name] = experiment_metrics_set

        for experiment_metrics_set in experiment_metrics_sets.values():
            experiment_metrics_set.generate_experiment_summary()

        return {
            'experiment_metrics_sets': experiment_metrics_sets
        }
    
    @event('generate_metrics_sets')
    async def generate_system_metrics(
        self,
        analyze_stage_monitors: MonitorGroup={},
        session_stage_monitors: MonitorGroup={},
        analyze_stage_batch_sizes: Dict[str, int]={},
    ):
        
        session_stage_monitors.update({
            self.name: {
                **analyze_stage_monitors
            }
        })

        system_metrics_set = SystemMetricsSet(
            session_stage_monitors,
            analyze_stage_batch_sizes
        )
        system_metrics_set.generate_system_summaries()

        return {
            'analyze_stage_system_metrics': system_metrics_set
        }

    @context(
        'generate_experiment_metrics',
        'generate_system_metrics'
    )
    async def generate_summary(
        self,
        analyze_stage_stages_count: int=0,
        analyze_stage_total_group_results: int=0,
        analyze_stage_processed_results: ProcessedResultsSet=[],
        analyze_stage_contexts: Dict[str, Any]={},
        experiment_metrics_sets: Dict[str, ExperimentMetricsSet]={},
        analyze_stage_system_metrics: SystemMetricsSet=None,
        analyze_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]] = {}
    ):

        self.context[self.name] = analyze_stage_contexts
        
        summaries: Dict[
            str, 
            Union[
                str,
                Dict[str, StageMetricsSummary], 
                Dict[str, MetricsSet], 
                SystemMetricsSet
            ]
        ] = {
            'stages': {},
            'source': self.name,
            'experiment_metrics_sets': experiment_metrics_sets,
            'system_metrics': analyze_stage_system_metrics
        }

        for stage_name, stage_metrics in analyze_stage_processed_results:
            summaries['stages'][stage_name] = stage_metrics

        self.analysis_execution_time = round(
            time.monotonic() - self.analysis_execution_time_start
        )

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed results analysis for - {analyze_stage_stages_count} - stages in - {self.analysis_execution_time} seconds')
        await self.logger.spinner.set_default_message(f'Completed results analysis for {analyze_stage_total_group_results} actions and {analyze_stage_stages_count} stages over {self.analysis_execution_time} seconds')

        self.executor.shutdown()
        
        return {
            'analyze_stage_summary_metrics': summaries,
            'analyze_stage_monitors': {
                self.name: {
                    **analyze_stage_monitors
                }
            }
        }

    @event('generate_summary')
    async def complete(self):
        return {}



from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import (
    BaseModel,
    StrictStr
)


class HARConnectorConfig(BaseModel):
    filepath: StrictStr
    connector_type: ConnectorType=ConnectorType.CSV